from dataclasses import dataclass
from typing import List, Optional, TypedDict

from ..utils.enums import ImageType, MediaType


@dataclass
class LuxonisComponent:
    """Abstraction for a piece of media within a source. Most commonly,
    this abstracts an image sensor.

    @type name: str
    @param name: A recognizable name for the component.
    @type media_type: L{MediaType}
    @param media_type: Enum for the type of media for the component.
    @type image_type: L{ImageType}
    @param image_type: Enum for the image type. Only used if
        C{media_type==MediaType.IMAGE}.
    """

    name: str
    media_type: MediaType = MediaType.IMAGE
    image_type: ImageType = ImageType.COLOR

    class LuxonisComponentDocument(TypedDict):
        name: str
        media_type: str
        image_type: str

    def to_document(self) -> LuxonisComponentDocument:
        return {
            "name": self.name,
            "media_type": self.media_type.value,
            "image_type": self.image_type.value,
        }

    @classmethod
    def from_document(
        cls, document: LuxonisComponentDocument
    ) -> "LuxonisComponent":
        if document["image_type"] is not None:
            return cls(
                name=document["name"],
                media_type=MediaType(document["media_type"]),
                image_type=ImageType(document["image_type"]),
            )
        else:
            return cls(
                name=document["name"],
                media_type=MediaType(document["media_type"]),
            )


class LuxonisSource:
    """Abstracts the structure of a dataset and which components/media
    are included."""

    class LuxonisSourceDocument(TypedDict):
        name: str
        main_component: str
        components: List[LuxonisComponent.LuxonisComponentDocument]

    def __init__(
        self,
        name: str = "default",
        components: Optional[List[LuxonisComponent]] = None,
        main_component: Optional[str] = None,
    ) -> None:
        """Abstracts the structure of a dataset by grouping together
        components.

        For example, with an U{OAK-D<https://docs.luxonis.com/projects/hardware
        /en/latest/pages/BW1098OAK/>}, you can have a source with 4 image
        components: C{rgb} (color), C{left} (mono), C{right} (mono), and C{depth}.

        @type name: str
        @param name: A recognizable name for the source. Defaults to "default".

        @type components: Optional[List[LuxonisComponent]]
        @param components: If not using the default configuration, a list of
        L{LuxonisComponent} to group together in the source.

        @type main_component: Optional[str]
        @param main_component: The name of the component that should be
            primarily visualized.
        """

        self.name = name
        if components is None:
            components = [
                LuxonisComponent(name)
            ]  # basic source includes a single color image

        self.components = {
            component.name: component for component in components
        }
        self.main_component = main_component or next(iter(self.components))

    def to_document(self) -> LuxonisSourceDocument:
        return {
            "name": self.name,
            "main_component": self.main_component,
            "components": [
                component.to_document()
                for component in self.components.values()
            ],
        }

    @classmethod
    def from_document(cls, document: LuxonisSourceDocument) -> "LuxonisSource":
        return cls(
            name=document.get("name", "default"),
            main_component=document.get("main_component"),
            components=[
                LuxonisComponent.from_document(component_doc)
                for component_doc in document["components"]
            ]
            if "components" in document
            else None,
        )



from typing import Any, Dict, List, Optional, Tuple

import albumentations as A
import cv2
import numpy as np

from luxonis_ml.utils.registry import Registry

from ..utils.enums import LabelType
from .batch_compose import BatchCompose
from .batch_transform import BatchBasedTransform

AUGMENTATIONS = Registry(name="augmentations")


class Augmentations:
    def __init__(
        self,
        image_size: List[int],
        augmentations: List[Dict[str, Any]],
        train_rgb: bool = True,
        keep_aspect_ratio: bool = True,
        only_normalize: bool = False,
    ):
        """Base class for augmentations that are used in LuxonisLoader.

        @type train_rgb: bool
        @param train_rgb: Whether should use RGB or BGR images.
        """

        self.image_size = image_size
        self.train_rgb = train_rgb
        self.only_normalize = only_normalize

        self.is_batched = False
        self.aug_batch_size = 1

        (
            self.batch_transform,
            self.spatial_transform,
            self.pixel_transform,
            self.resize_transform,
        ) = self._parse_cfg(
            image_size=image_size,
            augmentations=[
                a for a in augmentations if a["name"] == "Normalize"
            ]
            if only_normalize
            else augmentations,
            keep_aspect_ratio=keep_aspect_ratio,
        )

    def _parse_cfg(
        self,
        image_size: List[int],
        augmentations: List[Dict[str, Any]],
        keep_aspect_ratio: bool = True,
    ) -> Tuple[BatchCompose, A.Compose, A.Compose, A.Compose]:
        """Parses provided config and returns Albumentations
        BatchedCompose object and Compose object for default transforms.

        @type image_size: List[int]
        @param image_size: Desired image size [H,W]
        @type augmentations: List[Dict[str, Any]]
        @param augmentations: List of augmentations to use and their
            params
        @type keep_aspect_ratio: bool
        @param keep_aspect_ratio: Whether should use resize that keeps
            aspect ratio of original image.
        @rtype: Tuple[BatchCompose, A.Compose, A.Compose, A.Compose]
        @return: Objects for batched, spatial, pixel and resize
            transforms
        """

        # NOTE: Always perform Resize
        if keep_aspect_ratio:
            resize = AUGMENTATIONS.get("LetterboxResize")(
                height=image_size[0], width=image_size[1]
            )
        else:
            resize = A.Resize(image_size[0], image_size[1])

        pixel_augs = []
        spatial_augs = []
        batched_augs = []
        if augmentations:
            for aug in augmentations:
                curr_aug = AUGMENTATIONS.get(aug["name"])(
                    **aug.get("params", {})
                )
                if isinstance(curr_aug, A.ImageOnlyTransform):
                    pixel_augs.append(curr_aug)
                elif isinstance(curr_aug, A.DualTransform):
                    spatial_augs.append(curr_aug)
                elif isinstance(curr_aug, BatchBasedTransform):
                    self.is_batched = True
                    self.aug_batch_size = max(
                        self.aug_batch_size, curr_aug.batch_size
                    )
                    batched_augs.append(curr_aug)

        batch_transform = BatchCompose(
            [
                *batched_augs,
            ],
            bbox_params=A.BboxParams(
                format="coco",
                label_fields=[
                    "bboxes_classes_batch",
                    "bboxes_visibility_batch",
                ],
            ),
            keypoint_params=A.KeypointParams(
                format="xy",
                label_fields=[
                    "keypoints_visibility_batch",
                    "keypoints_classes_batch",
                ],
                remove_invisible=False,
            ),
        )

        spatial_transform = A.Compose(
            spatial_augs,
            bbox_params=A.BboxParams(
                format="coco",
                label_fields=["bboxes_classes", "bboxes_visibility"],
            ),
            keypoint_params=A.KeypointParams(
                format="xy",
                label_fields=["keypoints_visibility", "keypoints_classes"],
                remove_invisible=False,
            ),
        )

        pixel_transform = A.Compose(
            pixel_augs,
            bbox_params=A.BboxParams(
                format="coco",
                label_fields=["bboxes_classes", "bboxes_visibility"],
            ),
            keypoint_params=A.KeypointParams(
                format="xy",
                label_fields=["keypoints_visibility", "keypoints_classes"],
                remove_invisible=False,
            ),
        )

        resize_transform = A.Compose(
            [resize],
            bbox_params=A.BboxParams(
                format="coco",
                label_fields=["bboxes_classes", "bboxes_visibility"],
            ),
            keypoint_params=A.KeypointParams(
                format="xy",
                label_fields=["keypoints_visibility", "keypoints_classes"],
                remove_invisible=False,
            ),
        )

        return (
            batch_transform,
            spatial_transform,
            pixel_transform,
            resize_transform,
        )

    def _apply_transform(
        self,
        transformed: Dict[str, np.ndarray],
        arg_names: List[str],
        transform_func: A.Compose,
        return_mask: bool = False,
        arg_suffix: str = "",
    ) -> Dict[str, np.ndarray]:
        """Apply transform function.

        @type transformed: Dict[str, np.ndarray]
        @param transformed: Transformed data
        @type arg_names: List[str]
        @param arg_names: Names of arguments to pass to transform
            function
        @type transform_func: Callable
        @param transform_func: Transform function to apply
        @type return_mask: bool
        @param return_mask: Whether to return mask
        @type arg_suffix: str
        @param arg_suffix: Suffix to add to argument names
        @rtype: Dict[str, np.ndarray]
        @return: Transformed data
        """

        transform_args = {
            arg_name: transformed[f"{arg_name}{arg_suffix}"]
            for arg_name in arg_names
        }
        if return_mask:
            transform_args["mask"] = transformed[f"mask{arg_suffix}"]

        transformed = transform_func(force_apply=False, **transform_args)

        return transformed

    def __call__(
        self,
        data: List[Tuple[np.ndarray, Dict[LabelType, np.ndarray]]],
        ns: int = 1,
        nk: int = 1,
    ) -> Tuple[np.ndarray, Dict[LabelType, np.ndarray]]:
        """Performs augmentations on provided data.

        @type data: List[Tuple[np.ndarray, Dict[LabelType, np.ndarray]]]
        @param data: Data with list of input images and their
            annotations
        @type ns: int
        @param ns: Number of segmentation classes
        @type nk: int
        @param nk: Number of keypoints per instance
        @rtype: Tuple[np.ndarray, Dict[LabelType, np.ndarray]]
        @return: Output image and its annotations
        """

        present_annotations = {
            key for _, annotations in data for key in annotations.keys()
        }
        return_mask = LabelType.SEGMENTATION in present_annotations
        image_batch = []
        mask_batch = []
        bboxes_batch = []
        bboxes_visibility_batch = []
        bboxes_classes_batch = []
        keypoints_batch = []
        keypoints_visibility_batch = []
        keypoints_classes_batch = []

        bbox_counter = 0
        for img, annotations in data:
            (
                classes,
                mask,
                bboxes_points,
                bboxes_classes,
                keypoints_points,
                keypoints_visibility,
                keypoints_classes,
            ) = self.prepare_img_annotations(
                annotations, *img.shape[:-1], nk=nk, return_mask=return_mask
            )

            image_batch.append(img)
            if return_mask:
                mask_batch.append(mask)

            bboxes_batch.append(bboxes_points)
            bboxes_visibility_batch.append(
                [i + bbox_counter for i in range(bboxes_points.shape[0])]
            )
            bboxes_classes_batch.append(bboxes_classes)
            bbox_counter += bboxes_points.shape[0]

            keypoints_batch.append(keypoints_points)
            keypoints_visibility_batch.append(keypoints_visibility)
            keypoints_classes_batch.append(keypoints_classes)

        # Apply batch transform
        transform_args = {
            "image_batch": image_batch,
            "bboxes_batch": bboxes_batch,
            "bboxes_visibility_batch": bboxes_visibility_batch,
            "bboxes_classes_batch": bboxes_classes_batch,
            "keypoints_batch": keypoints_batch,
            "keypoints_visibility_batch": keypoints_visibility_batch,
            "keypoints_classes_batch": keypoints_classes_batch,
        }
        if return_mask:
            transform_args["mask_batch"] = mask_batch

        transformed = self.batch_transform(force_apply=False, **transform_args)
        transformed = {
            key: np.array(value[0]) for key, value in transformed.items()
        }

        arg_names = [
            "image",
            "bboxes",
            "bboxes_visibility",
            "bboxes_classes",
            "keypoints",
            "keypoints_visibility",
            "keypoints_classes",
        ]

        # Apply spatial transform
        transformed = self._apply_transform(
            transformed,
            arg_names,
            self.spatial_transform,
            return_mask,
            "_batch",
        )

        # Resize if necessary
        if transformed["image"].shape[:2] != self.image_size:
            transformed = self._apply_transform(
                transformed, arg_names, self.resize_transform, return_mask
            )

        # Apply pixel transform
        transformed = self._apply_transform(
            transformed, arg_names, self.pixel_transform, return_mask
        )

        out_image, out_mask, out_bboxes, out_keypoints = (
            self.post_transform_process(
                transformed,
                ns=ns,
                nk=nk,
                filter_kpts_by_bbox=(
                    LabelType.BOUNDINGBOX in present_annotations
                )
                and (LabelType.KEYPOINTS in present_annotations),
                return_mask=return_mask,
            )
        )

        out_annotations = {}
        for key in present_annotations:
            if key == LabelType.CLASSIFICATION:
                out_annotations[LabelType.CLASSIFICATION] = classes  # type: ignore
            elif key == LabelType.SEGMENTATION:
                out_annotations[LabelType.SEGMENTATION] = out_mask
            elif key == LabelType.BOUNDINGBOX:
                out_annotations[LabelType.BOUNDINGBOX] = out_bboxes
            elif key == LabelType.KEYPOINTS:
                out_annotations[LabelType.KEYPOINTS] = out_keypoints

        return out_image, out_annotations

    def prepare_img_annotations(
        self,
        annotations: Dict[LabelType, np.ndarray],
        ih: int,
        iw: int,
        nk: int,
        return_mask: bool = True,
    ) -> Tuple[
        np.ndarray,
        Optional[np.ndarray],
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]:
        """Prepare annotations to be compatible with albumentations.

        @type annotations: Dict[LabelType, np.ndarray]
        @param annotations: Dict with annotations
        @type ih: int
        @param ih: Input image height
        @type iw: int
        @param iw: Input image width
        @type return_mask: bool
        @param return_mask: Whether to compute and return mask
        @rtype: Tuple[np.ndarray, Optional[np.ndarray], np.ndarray,
            np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        @return: Annotations in albumentations format
        """

        classes = annotations.get(LabelType.CLASSIFICATION, np.zeros(1))

        mask = None
        if return_mask:
            seg = annotations.get(
                LabelType.SEGMENTATION, np.zeros((1, ih, iw))
            )
            mask = np.argmax(seg, axis=0) + 1
            mask[np.sum(seg, axis=0) == 0] = 0  # only background has value 0

        # COCO format in albumentations is [x,y,w,h] non-normalized
        bboxes = annotations.get(LabelType.BOUNDINGBOX, np.zeros((0, 5)))
        bboxes_points = bboxes[:, 1:]
        bboxes_points[:, 0::2] *= iw
        bboxes_points[:, 1::2] *= ih
        bboxes_points = self.check_bboxes(bboxes_points)
        bboxes_classes = bboxes[:, 0]

        # albumentations expects list of keypoints e.g. [(x,y),(x,y),(x,y),(x,y)]
        keypoints = annotations.get(
            LabelType.KEYPOINTS, np.zeros((1, nk * 3 + 1))
        )
        keypoints_unflat = np.reshape(keypoints[:, 1:], (-1, 3))
        keypoints_points = keypoints_unflat[:, :2]
        keypoints_points[:, 0] *= iw
        keypoints_points[:, 1] *= ih
        keypoints_visibility = keypoints_unflat[:, 2]
        # albumentations expects classes to be same length as keypoints
        # (use case: each kpt separate class - not supported in LuxonisDataset)
        keypoints_classes = np.repeat(keypoints[:, 0], nk)

        return (
            classes,
            mask,
            bboxes_points,
            bboxes_classes,
            keypoints_points,
            keypoints_visibility,
            keypoints_classes,
        )

    def post_transform_process(
        self,
        transformed_data: Dict[str, np.ndarray],
        ns: int,
        nk: int,
        filter_kpts_by_bbox: bool,
        return_mask: bool = True,
    ) -> Tuple[np.ndarray, Optional[np.ndarray], np.ndarray, np.ndarray]:
        """Postprocessing of albumentations output to LuxonisLoader
        format.

        @type transformed_data: Dict[str, np.ndarray]
        @param transformed_data: Output data from albumentations
        @type ns: int
        @param ns: Number of segmentation classes
        @type nk: int
        @param nk: Number of keypoints per instance
        @type filter_kpts_by_bbox: bool
        @param filter_kpts_by_bbox: If True removes keypoint instances
            if its bounding box was removed.
        @rtype: Tuple[np.ndarray, Optional[np.ndarray], np.ndarray,
            np.ndarray]
        @return: Postprocessed annotations
        """

        out_image = transformed_data["image"]
        ih, iw, _ = out_image.shape
        if not self.train_rgb:
            out_image = cv2.cvtColor(out_image, cv2.COLOR_RGB2BGR)
        out_image = out_image.astype(np.float32)

        out_mask = None
        if return_mask:
            transformed_mask = transformed_data.get("mask")
            out_mask = np.zeros((ns, *transformed_mask.shape))
            for key in np.unique(transformed_mask):
                if key != 0:
                    out_mask[int(key) - 1, ...] = transformed_mask == key
            out_mask[out_mask > 0] = 1

        if transformed_data["bboxes"]:
            transformed_bboxes_classes = np.expand_dims(
                transformed_data["bboxes_classes"], axis=-1
            )
            out_bboxes = np.concatenate(
                (transformed_bboxes_classes, transformed_data["bboxes"]),
                axis=1,
            )
        else:  # if no bboxes after transform
            out_bboxes = np.zeros((0, 5))
        out_bboxes[:, 1::2] /= iw
        out_bboxes[:, 2::2] /= ih

        transformed_keypoints_vis = np.expand_dims(
            transformed_data["keypoints_visibility"], axis=-1
        )

        if nk == 0:
            nk = 1  # done for easier postprocessing
        if transformed_data["keypoints"]:
            out_keypoints = np.concatenate(
                (transformed_data["keypoints"], transformed_keypoints_vis),
                axis=1,
            )
        else:
            out_keypoints = np.zeros((0, nk * 3 + 1))

        out_keypoints = self.mark_invisible_keypoints(out_keypoints, ih, iw)
        out_keypoints[..., 0] /= iw
        out_keypoints[..., 1] /= ih
        out_keypoints = np.reshape(out_keypoints, (-1, nk * 3))
        keypoints_classes = transformed_data["keypoints_classes"]
        keypoints_classes = keypoints_classes[0::nk]
        keypoints_classes = np.expand_dims(keypoints_classes, axis=-1)
        out_keypoints = np.concatenate(
            (keypoints_classes, out_keypoints), axis=1
        )
        if filter_kpts_by_bbox:
            out_keypoints = out_keypoints[
                transformed_data["bboxes_visibility"]
            ]  # keep only keypoints of visible instances

        return out_image, out_mask, out_bboxes, out_keypoints

    def check_bboxes(self, bboxes: np.ndarray) -> np.ndarray:
        """Check bbox annotations and correct those with width or height
        0.

        @type bboxes: np.ndarray
        @param bboxes: A numpy array representing bounding boxes.
        @rtype: np.ndarray
        @return: The same bounding boxes with any out-of-bounds
            coordinate corrections.
        """

        for i in range(bboxes.shape[0]):
            if bboxes[i, 2] == 0:
                bboxes[i, 2] = 1
            if bboxes[i, 3] == 0:
                bboxes[i, 3] = 1
        return bboxes

    def mark_invisible_keypoints(
        self, keypoints: np.ndarray, ih: int, iw: int
    ) -> np.ndarray:
        """Mark invisible keypoints with label == 0.

        @type keypoints: np.ndarray
        @param keypoints: A numpy array representing keypoints.
        @type ih: int
        @param ih: The image height.
        @type iw: int
        @param iw: The image width.
        @rtype: np.ndarray
        @return: The same keypoints with corrections to mark keypoints
            out-of-bounds as invisible.
        """
        for kp in keypoints:
            if not (0 <= kp[0] < iw and 0 <= kp[1] < ih):
                kp[2] = 0
            if kp[2] == 0:  # per COCO format invisible points have x=y=0
                kp[0] = kp[1] = 0
        return keypoints


# Registering all supported transforms
# Pixel-level transforms
AUGMENTATIONS.register_module(module=A.AdvancedBlur)
AUGMENTATIONS.register_module(module=A.Blur)
AUGMENTATIONS.register_module(module=A.CLAHE)
AUGMENTATIONS.register_module(module=A.ChannelDropout)
AUGMENTATIONS.register_module(module=A.ChannelShuffle)
AUGMENTATIONS.register_module(module=A.ColorJitter)
AUGMENTATIONS.register_module(module=A.Defocus)
AUGMENTATIONS.register_module(module=A.Downscale)
AUGMENTATIONS.register_module(module=A.Emboss)
AUGMENTATIONS.register_module(module=A.Equalize)
AUGMENTATIONS.register_module(module=A.FDA)
AUGMENTATIONS.register_module(module=A.FancyPCA)
AUGMENTATIONS.register_module(module=A.FromFloat)
AUGMENTATIONS.register_module(module=A.GaussNoise)
AUGMENTATIONS.register_module(module=A.GaussianBlur)
AUGMENTATIONS.register_module(module=A.GlassBlur)
AUGMENTATIONS.register_module(module=A.HistogramMatching)
AUGMENTATIONS.register_module(module=A.HueSaturationValue)
AUGMENTATIONS.register_module(module=A.ISONoise)
AUGMENTATIONS.register_module(module=A.ImageCompression)
AUGMENTATIONS.register_module(module=A.InvertImg)
AUGMENTATIONS.register_module(module=A.MedianBlur)
AUGMENTATIONS.register_module(module=A.MotionBlur)
AUGMENTATIONS.register_module(module=A.MultiplicativeNoise)
AUGMENTATIONS.register_module(module=A.Normalize)
AUGMENTATIONS.register_module(module=A.PixelDistributionAdaptation)
AUGMENTATIONS.register_module(module=A.Posterize)
AUGMENTATIONS.register_module(module=A.RGBShift)
AUGMENTATIONS.register_module(module=A.RandomBrightnessContrast)
AUGMENTATIONS.register_module(module=A.RandomFog)
AUGMENTATIONS.register_module(module=A.RandomGamma)
AUGMENTATIONS.register_module(module=A.RandomGravel)
AUGMENTATIONS.register_module(module=A.RandomRain)
AUGMENTATIONS.register_module(module=A.RandomShadow)
AUGMENTATIONS.register_module(module=A.RandomSnow)
AUGMENTATIONS.register_module(module=A.RandomSunFlare)
AUGMENTATIONS.register_module(module=A.RandomToneCurve)
AUGMENTATIONS.register_module(module=A.RingingOvershoot)
AUGMENTATIONS.register_module(module=A.Sharpen)
AUGMENTATIONS.register_module(module=A.Solarize)
AUGMENTATIONS.register_module(module=A.Spatter)
AUGMENTATIONS.register_module(module=A.Superpixels)
AUGMENTATIONS.register_module(module=A.TemplateTransform)
AUGMENTATIONS.register_module(module=A.ToFloat)
AUGMENTATIONS.register_module(module=A.ToGray)
AUGMENTATIONS.register_module(module=A.ToRGB)
AUGMENTATIONS.register_module(module=A.ToSepia)
AUGMENTATIONS.register_module(module=A.UnsharpMask)
AUGMENTATIONS.register_module(module=A.ZoomBlur)

# Spatial.level transforms
# NOTE: only augmentations that are supported for all targets
AUGMENTATIONS.register_module(module=A.Affine)
# AUGMENTATIONS.register_module(module=A.BBoxSafeRandomCrop)
AUGMENTATIONS.register_module(module=A.CenterCrop)
AUGMENTATIONS.register_module(module=A.CoarseDropout)
AUGMENTATIONS.register_module(module=A.Crop)
AUGMENTATIONS.register_module(module=A.CropAndPad)
AUGMENTATIONS.register_module(module=A.CropNonEmptyMaskIfExists)
# AUGMENTATIONS.register_module(module=A.ElasticTransform)
AUGMENTATIONS.register_module(module=A.Flip)
# AUGMENTATIONS.register_module(module=A.GridDistortion)
# AUGMENTATIONS.register_module(module=A.GridDropout)
AUGMENTATIONS.register_module(module=A.HorizontalFlip)
AUGMENTATIONS.register_module(module=A.Lambda)
AUGMENTATIONS.register_module(module=A.LongestMaxSize)
# AUGMENTATIONS.register_module(module=A.MaskDropout)
AUGMENTATIONS.register_module(module=A.NoOp)
# AUGMENTATIONS.register_module(module=A.OpticalDistortion)
AUGMENTATIONS.register_module(module=A.PadIfNeeded)
AUGMENTATIONS.register_module(module=A.Perspective)
AUGMENTATIONS.register_module(module=A.PiecewiseAffine)
AUGMENTATIONS.register_module(module=A.PixelDropout)
AUGMENTATIONS.register_module(module=A.RandomCrop)
AUGMENTATIONS.register_module(module=A.RandomCropFromBorders)
AUGMENTATIONS.register_module(module=A.RandomCropNearBBox)
AUGMENTATIONS.register_module(module=A.RandomGridShuffle)
AUGMENTATIONS.register_module(module=A.RandomResizedCrop)
AUGMENTATIONS.register_module(module=A.RandomRotate90)
AUGMENTATIONS.register_module(module=A.RandomScale)
# AUGMENTATIONS.register_module(module=A.RandomSizedBBoxSafeCrop)
AUGMENTATIONS.register_module(module=A.RandomSizedCrop)
AUGMENTATIONS.register_module(module=A.Resize)
AUGMENTATIONS.register_module(module=A.Rotate)
AUGMENTATIONS.register_module(module=A.SafeRotate)
AUGMENTATIONS.register_module(module=A.ShiftScaleRotate)
AUGMENTATIONS.register_module(module=A.SmallestMaxSize)
AUGMENTATIONS.register_module(module=A.Transpose)
AUGMENTATIONS.register_module(module=A.VerticalFlip)



from typing import Final

import numpy as np

from luxonis_ml.data.augmentations.custom.mixup import MixUp

WIDTH: Final[int] = 640
HEIGHT: Final[int] = 480


def test_mixup():
    img = (np.random.rand(HEIGHT, WIDTH, 3) * 255).astype(np.uint8)
    mixup = MixUp(p=1.0, always_apply=True)
    m = mixup.apply_to_image_batch(
        image_batch=[img, img], image_shapes=[(HEIGHT, WIDTH), (HEIGHT, WIDTH)]
    )
    assert m[0].shape == (HEIGHT, WIDTH, 3)
    result = mixup(image_batch=[img, img], labels={})
    assert result["image_batch"][0].shape == (HEIGHT, WIDTH, 3)



from enum import Enum


class LabelType(str, Enum):
    """Supported computer vision label types."""

    CLASSIFICATION = "classification"
    SEGMENTATION = "segmentation"
    BOUNDINGBOX = "boundingbox"
    KEYPOINTS = "keypoints"
    LABEL = "label"
    ARRAY = "array"


class DataLabelType(Enum):
    """Supported computer vision label types.

    Annotation types can be nested (I{e.g.} a BOX has 2 LABELS, a BOX
    has a POLYLINE instance segmentation, I{etc.})
    """

    CLASSIFICATION = "classification"  # used for single, multi-class, or multi-label classification
    BOX = "box"  # bounding box
    POLYLINE = "polyline"  # polyline to represent segmentation mask instances
    SEGMENTATION = "segmentation"  # RLE encoding of a binary segmentation mask
    KEYPOINTS = "keypoints"  # keypoint skeleton instances
    LABEL = "label"  # an arbitrary label of string, bool, or number
    ARRAY = "array"  # a path to a numpy (.npy) array


class MediaType(Enum):
    """Individual file type."""

    IMAGE = "image"
    VIDEO = "video"
    POINTCLOUD = "pointcloud"


class ImageType(Enum):
    """Image type for IMAGE HType."""

    COLOR = "color"  # 3 channel BGR or RGB (uint8)
    MONO = "mono"  # 1 channel (uint8)
    DISPARITY = "disparity"  # disparity or depth (uint16)


class BucketType(Enum):
    """Whether bucket storage is internal to Luxonis or not."""

    INTERNAL = "internal"
    EXTERNAL = "external"


class BucketStorage(Enum):
    """Underlying object storage for a bucket."""

    LOCAL = "local"
    S3 = "s3"
    GCS = "gcs"
    AZURE_BLOB = "azure"



from pathlib import Path
from typing import Any, Dict

import polars as pl

from luxonis_ml.utils.filesystem import PathType


class ParquetFileManager:
    def __init__(self, directory: PathType, num_rows: int = 100_000) -> None:
        """Manages the insertion of data into parquet files.

        @type directory: str
        @param directory: The local directory in which parquet files are
            stored.
        @type num_rows: int
        @param num_rows: The maximum number of rows permitted in a
            parquet file before another file is created.
        """

        self.dir = Path(directory)
        self.parquet_files = list(self.dir.glob("*.parquet"))
        self.num_rows = num_rows

        self.num = self._find_num() if self.parquet_files else 0
        self.current_file = self._generate_filename(self.num)

        self._read()

    def _find_num(self) -> int:
        return max(int(file.stem) for file in self.parquet_files)

    def _generate_filename(self, num: int) -> Path:
        return self.dir / f"{num:010d}.parquet"

    def _read(self) -> None:
        if self.current_file.exists():
            df = pl.read_parquet(self.current_file)
            self.buffer = df.to_dict(as_series=False)
            self.row_count = len(df)
        else:
            self.row_count = 0
            self.buffer = {}

    def _initialize_data(self, data: Dict) -> None:
        for key in data:
            self.buffer[key] = []

    def write(self, add_data: Dict[str, Any]) -> None:
        """Writes a row to the current working parquet file.

        @type add_data: Dict
        @param add_data: A dictionary representing annotations, mapping
            annotation types to values.
        """

        if not self.buffer:
            self._initialize_data(add_data)

        for key in add_data:
            if key not in self.buffer:
                raise KeyError(f"Key {key} Not Found")
            self.buffer[key].append(add_data[key])

        self.row_count += 1
        if self.row_count % self.num_rows == 0:
            self._flush()
            self.num += 1
            self.current_file = self._generate_filename(self.num)
            self._read()

    def _flush(self) -> None:
        """Writes buffered data to parquet."""

        if self.buffer:
            df = pl.DataFrame(self.buffer).cast({"instance_id": pl.Int16})
            df.write_parquet(self.current_file)

    def __enter__(self) -> "ParquetFileManager":
        return self

    def __exit__(self, *_) -> None:
        self._flush()



from pathlib import Path
from typing import Dict, Iterator, Tuple

import numpy as np


def check_array(path: Path) -> None:
    """Checks whether a path to a numpy array is valid. This checks that
    th file exists and is readable by numpy.

    @type values: Path
    @param values: A path to a numpy array.
    @rtype: NoneType
    @return: None
    @raise ValueError: If the path is not a valid numpy array.
    """

    def _check_valid_array(path: Path) -> bool:
        try:
            np.load(path)
            return True
        except Exception:
            return False

    if not isinstance(path, Path) or not path.suffix == ".npy":
        raise ValueError(
            f"Array path {path} must be a path to a numpy array (.npy)"
        )
    if not _check_valid_array(path):
        raise ValueError(
            f"Array at path {path} is not a valid numpy array (.npy)"
        )


def rgb_to_bool_masks(
    segmentation_mask: np.ndarray,
    class_colors: Dict[str, Tuple[int, int, int]],
    add_background_class: bool = False,
) -> Iterator[Tuple[str, np.ndarray]]:
    """Helper function to convert an RGB segmentation mask to boolean
    masks for each class.

    Example::
        >>> segmentation_mask = np.array([[[0, 0, 0], [255, 0, 0], [0, 255, 0]],
        ...                                [[0, 0, 0], [0, 255, 0], [0, 0, 255]]], dtype=np.uint8)
        >>> class_colors = {"red": (255, 0, 0), "green": (0, 255, 0), "blue": (0, 0, 255)}
        >>> for class_name, mask in rgb_to_bool_masks(segmentation_mask, class_colors, add_background_class=True):
        ...     print(class_name, mask)
        background [[ True False False],
                    [ True False False]]
        red        [[False  True False],
                    [False False False]]
        green      [[False False  True],
                    [False True False]]
        blue       [[False False False],
                    [False False  True]]

    @type segmentation_mask: npt.NDArray[np.uint8]
    @param segmentation_mask: An RGB segmentation mask where each pixel is colored according to the class it belongs to.
    @type class_colors: Dict[str, Tuple[int, int, int]]
    @param class_colors: A dictionary mapping class names to RGB colors.
    @type add_background_class: bool
    @param add_background_class: Whether to add a background class with a mask for all pixels
        that do not belong to any class. The class name will be set to "background".
        The background class will be yielded first. Default is False.
    @rtype: Iterator[Tuple[str, npt.NDArray[np.bool_]]]
    @return: An iterator of tuples where the first element is the class name and
        the second element is a boolean mask for that class.
    """
    color_to_id = {
        tuple(color): i for i, color in enumerate(class_colors.values())
    }

    lookup_table = np.zeros((256, 256, 256), dtype=np.uint8)
    for color, id in color_to_id.items():
        lookup_table[color[0], color[1], color[2]] = id + 1

    segmentation_ids = lookup_table[
        segmentation_mask[:, :, 0],
        segmentation_mask[:, :, 1],
        segmentation_mask[:, :, 2],
    ]

    if add_background_class:
        background_mask = segmentation_ids == 0
        yield "background", background_mask

    for class_name, color in class_colors.items():
        class_id = color_to_id[tuple(color)] + 1
        yield class_name, segmentation_ids == class_id



from .annotation import (
    Annotation,
    ArrayAnnotation,
    BBoxAnnotation,
    ClassificationAnnotation,
    DatasetRecord,
    KeypointAnnotation,
    LabelAnnotation,
    PolylineSegmentationAnnotation,
    RLESegmentationAnnotation,
    load_annotation,
)
from .base_dataset import DATASETS_REGISTRY, BaseDataset, DatasetIterator
from .luxonis_dataset import LuxonisDataset
from .source import LuxonisComponent, LuxonisSource

__all__ = [
    "BaseDataset",
    "DatasetIterator",
    "DatasetRecord",
    "LuxonisDataset",
    "LuxonisComponent",
    "LuxonisSource",
    "DATASETS_REGISTRY",
    "Annotation",
    "ClassificationAnnotation",
    "BBoxAnnotation",
    "KeypointAnnotation",
    "RLESegmentationAnnotation",
    "PolylineSegmentationAnnotation",
    "ArrayAnnotation",
    "LabelAnnotation",
    "load_annotation",
]



from typing import Any, Dict, Tuple

import cv2
import numpy as np
from albumentations import BoxType, DualTransform, KeypointType
from albumentations.core.bbox_utils import denormalize_bbox, normalize_bbox

from ..utils import AUGMENTATIONS


@AUGMENTATIONS.register_module()
class LetterboxResize(DualTransform):
    def __init__(
        self,
        height: int,
        width: int,
        interpolation: int = cv2.INTER_LINEAR,
        border_value: int = 0,
        mask_value: int = 0,
        always_apply: bool = False,
        p: float = 1.0,
    ):
        """Augmentation to apply letterbox resizing to images. Also
        transforms masks, bboxes and keypoints to correct shape.

        @type height: int
        @param height: Desired height of the output.
        @type width: int
        @param width: Desired width of the output.
        @type interpolation: int, optional
        @param interpolation: cv2 flag to specify interpolation used
            when resizing. Defaults to C{cv2.INTER_LINEAR}.
        @type border_value: int, optional
        @param border_value: Padding value for images. Defaults to C{0}.
        @type mask_value: int, optional
        @param mask_value: Padding value for masks. Defaults to C{0}.
        @type always_apply: bool, optional
        @param always_apply: Whether to always apply the transform.
            Defaults to C{False}.
        @type p: float, optional
        @param p: Probability of applying the transform. Defaults to
            C{1.0}.
        """

        super().__init__(always_apply, p)

        if not (0 <= border_value <= 255):
            raise ValueError("Border value must be in range [0,255].")

        if not (0 <= mask_value <= 255):
            raise ValueError("Mask value must be in range [0,255].")

        self.height = height
        self.width = width
        self.interpolation = interpolation
        self.border_value = border_value
        self.mask_value = mask_value

    def update_params(
        self, params: Dict[str, Any], **kwargs
    ) -> Dict[str, Any]:
        """Updates augmentation parameters with the necessary metadata.

        @param params: The existing augmentation parameters dictionary.
        @type params: Dict[str, Any]
        @param kwargs: Additional keyword arguments to add the
            parameters.
        @type kwargs: Any
        @return: Updated dictionary containing the merged parameters.
        @rtype: Dict[str, Any]
        """

        params = super().update_params(params, **kwargs)

        img_height = params["rows"]
        img_width = params["cols"]

        ratio = min(self.height / img_height, self.width / img_width)
        new_height = int(img_height * ratio)
        new_width = int(img_width * ratio)

        # only supports center alignment
        pad_top = (self.height - new_height) // 2
        pad_bottom = pad_top

        pad_left = (self.width - new_width) // 2
        pad_right = pad_left

        params.update(
            {
                "pad_top": pad_top,
                "pad_bottom": pad_bottom,
                "pad_left": pad_left,
                "pad_right": pad_right,
            }
        )

        return params

    def apply(
        self,
        img: np.ndarray,
        pad_top: int,
        pad_bottom: int,
        pad_left: int,
        pad_right: int,
        **kwargs,
    ) -> np.ndarray:
        """Applies the letterbox augmentation to an image.

        @type img: np.ndarray
        @param img: Input image to which resize is applied.
        @type pad_top: int
        @param pad_top: Number of pixels to pad at the top.
        @type pad_bottom: int
        @param pad_bottom: Number of pixels to pad at the bottom.
        @type pad_left: int
        @param pad_left: Number of pixels to pad on the left.
        @type pad_right: int
        @param pad_right: Number of pixels to pad on the right.
        @type kwargs: Any
        @param kwargs: Additional parameters for the padding operation.
        @rtype: np.ndarray
        @return: Image with applied letterbox resize.
        """

        resized_img = cv2.resize(
            img,
            (
                self.width - pad_left - pad_right,
                self.height - pad_top - pad_bottom,
            ),
            interpolation=self.interpolation,
        )
        img_out = cv2.copyMakeBorder(
            resized_img,
            pad_top,
            pad_bottom,
            pad_left,
            pad_right,
            cv2.BORDER_CONSTANT,
            self.border_value,
        )
        img_out = img_out.astype(img.dtype)
        return img_out

    def apply_to_mask(
        self,
        img: np.ndarray,
        pad_top: int,
        pad_bottom: int,
        pad_left: int,
        pad_right: int,
        **params,
    ) -> np.ndarray:
        """Applies letterbox augmentation to the input mask.

        @type img: np.ndarray
        @param img: Input mask to which resize is applied.
        @type pad_top: int
        @param pad_top: Number of pixels to pad at the top.
        @type pad_bottom: int
        @param pad_bottom: Number of pixels to pad at the bottom.
        @type pad_left: int
        @param pad_left: Number of pixels to pad on the left.
        @type pad_right: int
        @param pad_right: Number of pixels to pad on the right.
        @type params: Any
        @param params: Additional parameters for the padding operation.
        @rtype: np.ndarray
        @return: Mask with applied letterbox resize.
        """

        resized_img = cv2.resize(
            img,
            (
                self.width - pad_left - pad_right,
                self.height - pad_top - pad_bottom,
            ),
            interpolation=cv2.INTER_NEAREST,
        )
        img_out = cv2.copyMakeBorder(
            resized_img,
            pad_top,
            pad_bottom,
            pad_left,
            pad_right,
            cv2.BORDER_CONSTANT,
            self.mask_value,
        )
        img_out = img_out.astype(img.dtype)
        return img_out

    def apply_to_bbox(
        self,
        bbox: BoxType,
        pad_top: int,
        pad_bottom: int,
        pad_left: int,
        pad_right: int,
        **params,
    ) -> BoxType:
        """Applies letterbox augmentation to the bounding box.

        @type bbox: BoxType
        @param bbox: Bounding box to which resize is applied.
        @type pad_top: int
        @param pad_top: Number of pixels to pad at the top.
        @type pad_bottom: int
        @param pad_bottom: Number of pixels to pad at the bottom.
        @type pad_left: int
        @param pad_left: Number of pixels to pad on the left.
        @type pad_right: int
        @param pad_right: Number of pixels to pad on the right.
        @type params: Any
        @param params: Additional parameters for the padding operation.
        @rtype: BoxType
        @return: Bounding box with applied letterbox resize.
        """

        x_min, y_min, x_max, y_max = denormalize_bbox(
            bbox,
            self.height - pad_top - pad_bottom,
            self.width - pad_left - pad_right,
        )[:4]
        bbox = np.array(
            [
                x_min + pad_left,
                y_min + pad_top,
                x_max + pad_left,
                y_max + pad_top,
            ]
        )
        # clip bbox to image, ignoring padding
        bbox = bbox.clip(
            min=[pad_left, pad_top] * 2,
            max=[self.width - pad_left, self.height - pad_top] * 2,
        ).tolist()
        return normalize_bbox(bbox, self.height, self.width)

    def apply_to_keypoint(
        self,
        keypoint: KeypointType,
        pad_top: int,
        pad_bottom: int,
        pad_left: int,
        pad_right: int,
        **kwargs,
    ) -> KeypointType:
        """Applies letterbox augmentation to the keypoint.

        @type keypoint: KeypointType
        @param keypoint: Keypoint to which resize is applied.
        @type pad_top: int
        @param pad_top: Number of pixels to pad at the top.
        @type pad_bottom: int
        @param pad_bottom: Number of pixels to pad at the bottom.
        @type pad_left: int
        @param pad_left: Number of pixels to pad on the left.
        @type pad_right: int
        @param pad_right: Number of pixels to pad on the right.
        @type kwargs: Any
        @param kwargs: Additional parameters for the padding operation.
        @rtype: KeypointType
        @return: Keypoint with applied letterbox resize.
        """

        x, y, angle, scale = keypoint[:4]
        scale_x = (self.width - pad_left - pad_right) / kwargs["cols"]
        scale_y = (self.height - pad_top - pad_bottom) / kwargs["rows"]
        new_x = (x * scale_x) + pad_left
        new_y = (y * scale_y) + pad_top
        # if keypoint is in the padding then set coordinates to -1
        out_keypoint = (
            new_x
            if not self._out_of_bounds(new_x, pad_left, self.width - pad_left)
            else -1,
            new_y
            if not self._out_of_bounds(new_y, pad_top, self.height - pad_top)
            else -1,
            angle,
            scale * max(scale_x, scale_y),
        )
        return out_keypoint

    def get_transform_init_args_names(self) -> Tuple[str, ...]:
        """Gets the default arguments for the letterbox augmentation.

        @rtype: Tuple[str, ...]
        @return: The string keywords of the arguments.
        """

        return (
            "height",
            "width",
            "interpolation",
            "border_value",
            "mask_value",
        )

    def _out_of_bounds(
        self, value: float, min_limit: float, max_limit: float
    ) -> bool:
        """ "Check if the given value is outside the specified limits.

        @type value: float
        @param value: The value to be checked.
        @type min_limit: float
        @param min_limit: Minimum limit.
        @type max_limit: float
        @param max_limit: Maximum limit.
        @rtype: bool
        @return: True if the value is outside the specified limits,
            False otherwise.
        """
        return value < min_limit or value > max_limit



import platform
import shutil
import sys
from pathlib import Path

import pytest

from luxonis_ml.utils import setup_logging
from luxonis_ml.utils.environ import environ

setup_logging(use_rich=True, rich_print=True, configure_warnings=True)


@pytest.fixture(autouse=True, scope="module")
def set_paths():
    environ.LUXONISML_BASE_PATH = Path.cwd() / "tests/data/luxonisml_base_path"
    if environ.LUXONISML_BASE_PATH.exists():
        shutil.rmtree(environ.LUXONISML_BASE_PATH)


@pytest.fixture(scope="session")
def python_version():
    version = sys.version_info
    formatted_version = f"{version.major}{version.minor}"
    return formatted_version


@pytest.fixture(scope="session")
def platform_name():
    os_name = platform.system().lower()
    if "darwin" in os_name:
        return "mac"
    elif "linux" in os_name:
        return "lin"
    elif "windows" in os_name:
        return "win"
    else:
        raise ValueError(f"Unsupported operating system: {os_name}")



from typing import Any, Callable, Dict, List, Sequence

import numpy as np
from albumentations.core.transforms_interface import (
    BasicTransform,
    BoxType,
    KeypointType,
)


class BatchBasedTransform(BasicTransform):
    def __init__(self, batch_size: int, **kwargs):
        """Transform for multi-image.

        @param batch_size: Batch size needed for augmentation to work
        @type batch_size: int
        @param kwargs: Additional BasicTransform parameters
        @type kwargs: Any
        """
        super().__init__(**kwargs)

        self.batch_size = batch_size

    @property
    def targets(self) -> Dict[str, Callable]:
        return {
            "image_batch": self.apply_to_image_batch,
            "mask_batch": self.apply_to_mask_batch,
            "bboxes_batch": self.apply_to_bboxes_batch,
            "keypoints_batch": self.apply_to_keypoints_batch,
        }

    def update_params(
        self, params: Dict[str, Any], **kwargs
    ) -> Dict[str, Any]:
        # This overwrites the `super().update_params(...)`
        return params

    def apply_to_image_batch(
        self, image_batch: Sequence[BoxType], **params
    ) -> List[np.ndarray]:
        raise NotImplementedError(
            "Method apply_to_image_batch is not implemented in class "
            + self.__class__.__name__
        )

    def apply_to_mask_batch(
        self, mask_batch: Sequence[BoxType], **params
    ) -> List[np.ndarray]:
        raise NotImplementedError(
            "Method apply_to_mask_batch is not implemented in class "
            + self.__class__.__name__
        )

    def apply_to_bboxes_batch(
        self, bboxes_batch: Sequence[BoxType], **params
    ) -> List[BoxType]:
        raise NotImplementedError(
            "Method apply_to_bboxes_batch is not implemented in class "
            + self.__class__.__name__
        )

    def apply_to_keypoints_batch(
        self, keypoints_batch: Sequence[BoxType], **params
    ) -> List[KeypointType]:
        raise NotImplementedError(
            "Method apply_to_keypoints_batch is not implemented in class "
            + self.__class__.__name__
        )



"""Out-of-Distribution Detection for Embeddings.

This module provides two primary methods for detecting out-of-distribution (OOD) samples
in embeddings. OOD samples can be crucial to identify as they represent anomalies or novel
patterns that don't conform to the expected distribution of the dataset.

Methods available:
    - Isolation Forests: A tree-based model that partitions the space in such a manner that
      anomalies are isolated from the rest.

    - Leverage with Linear Regression: Leverages (or hat values) represent the distance between
      the predicted values and the true values. Higher leverages indicate potential OOD points.

Typical use cases include:
    - Anomaly Detection: Identifying rare patterns or outliers.

    - Dataset Reduction: By removing or studying OOD samples, we can have a more homogeneous dataset.

    - Expanding Datasets: Recognizing valuable data points that are distinct from the current distribution can be helpful
      when we're looking to diversify the dataset, especially in iterative learning scenarios.

Dependencies:
    - numpy
    - scikit-learn
"""

from typing import Optional, Union

import numpy as np
from sklearn.ensemble import IsolationForest


def isolation_forest_OOD(
    X: np.array,
    contamination: Union[float, str] = "auto",
    n_jobs: int = -1,
    verbose: int = 1,
    random_state: Optional[int] = None,
) -> np.array:
    """Out-of-distribution detection using Isolation Forests.

    @type X: np.array
    @param X: The embeddings to use.
    @type contamination: Union[float, str]
    @param contamination: The contamination parameter for Isolation
        Forests. Default is 'auto'.
    @type n_jobs: int
    @param n_jobs: The number of jobs to use. Default is -1, which means
        all available CPUs.
    @type verbose: int
    @param verbose: The verbosity level. Default is 1.
    @type random_state: Optional[int]
    @param random_state: The random state to use. Default is None.
    @rtype: np.array
    @return: The indices of the embeddings that are in-distribution.
    """
    # Initialize the Isolation Forest model
    isolation_forest = IsolationForest(
        contamination=contamination,
        n_jobs=n_jobs,
        verbose=verbose,
        random_state=random_state,
    )

    # Fit the model on all embeddings
    isolation_forest.fit(X)

    # Predict the outliers
    predicted_labels = isolation_forest.predict(X)

    # Get the indices of the outliers (where the predicted label is -1)
    outlier_indices_forest = np.where(predicted_labels == -1)[0]

    return outlier_indices_forest


def leverage_OOD(X: np.array, std_threshold: int = 3) -> np.array:
    """Out-of-distribution detection using leverage and linear
    regression.

    @type X: np.array
    @param X: The embeddings to use.
    @type std_threshold: int
    @param std_threshold: The number of standard deviations to use for
        the leverage threshold. Default is 3.
    @rtype: np.array
    @return: The indices of the embeddings that are out-of-distribution.
    """
    # Calculate the hat matrix (projection matrix) to get the leverage for each point
    hat_matrix = np.matmul(np.matmul(X, np.linalg.inv(np.matmul(X.T, X))), X.T)
    leverage = np.diagonal(hat_matrix)

    # Calculate the leverage threshold using 3 standard deviations
    mean, std = np.mean(leverage), np.std(leverage)
    upper_threshold, lower_threshold = (
        mean + std_threshold * std,
        mean - std_threshold * std,
    )
    outlier_indices_lev_3std = np.where(
        (leverage > upper_threshold) | (leverage < lower_threshold)
    )[0]

    return outlier_indices_lev_3std



from enum import Enum
from pathlib import Path
from typing import Optional

import typer
from rich import print
from typing_extensions import Annotated

from luxonis_ml.utils import LuxonisFileSystem

app = typer.Typer()

UrlArgument = Annotated[str, typer.Argument(..., help="URL of the file.")]


class TypeEnum(str, Enum):
    FILE = "file"
    DIR = "directory"
    ALL = "all"


@app.command()
def get(
    url: UrlArgument,
    save_dir: Annotated[
        Optional[Path], typer.Argument(help="Directory to save the file.")
    ] = None,
):
    """Downloads file from remote storage."""
    LuxonisFileSystem.download(url, save_dir or Path.cwd())


@app.command()
def put(
    file: Annotated[Path, typer.Argument(help="Path to the file to upload.")],
    url: UrlArgument,
):
    """Uploads file to remote storage."""
    LuxonisFileSystem.upload(file, url)


@app.command()
def delete(url: UrlArgument):
    """Deletes file from remote storage."""
    fs = LuxonisFileSystem(url)
    fs.delete_file("")


@app.command()
def ls(
    url: UrlArgument,
    recursive: Annotated[
        bool,
        typer.Option(..., "--recursive", "-r", help="List files recursively."),
    ] = False,
    typ: Annotated[
        TypeEnum,
        typer.Option(
            ...,
            "--type",
            "-t",
            help="Type of the files to list. If not provided, all files will be listed.",
        ),
    ] = TypeEnum.ALL,
):
    """Lists files in the remote directory."""
    fs = LuxonisFileSystem(url.rstrip("/"))
    for file in fs.walk_dir("", recursive=recursive, typ=typ.value):
        print(file)



"""Qdrant Docker Management and Embedding Operations.

This script provides utility functions for managing Qdrant via Docker and performing operations related to embeddings.

Features include:
    - Docker management: Checks for Docker installation, running status, and existence of specific images or containers.
    - Qdrant management: Facilitates starting a Qdrant container, connecting to Qdrant, and creating collections.
    - Embedding operations: Supports inserting, batch inserting, searching, and retrieving embeddings within a Qdrant collection.

Dependencies:
    - os: Standard library module for interacting with the operating system.
    - docker: Library to manage Docker containers.
    - qdrant_client: Client library for interacting with the Qdrant service.

Usage steps:
    1. Ensure Docker is installed and running on your system.
    2. Utilize the QdrantManager class and its start_docker_qdrant() method to initiate a Qdrant container.
    3. Employ the QdrantAPI class to either connect to an existing Qdrant service or create a new collection.
    4. Use the QdrantAPI class to perform various embedding-related operations within the specified Qdrant collection.

Note:
    - The default collection name is 'mnist'.
    - It is essential that the user has appropriate permissions to execute Docker commands without sudo.
For guidance on setting this up, refer to: https://docs.docker.com/engine/install/linux-postinstall/
"""

import os
from typing import Any, Dict, List, Optional, Tuple

import docker
from qdrant_client import QdrantClient
from qdrant_client.http import models
from qdrant_client.models import Distance, SearchRequest, VectorParams

from luxonis_ml.embeddings.utils.vectordb import VectorDBAPI


class QdrantManager:
    """Class to manage Qdrant Docker container and perform various
    operations related to embeddings."""

    def __init__(
        self, image_name="qdrant/qdrant", container_name="qdrant_container"
    ):
        """Initialize the QdrantManager."""
        self.image_name = image_name
        self.container_name = container_name
        self.client_docker = docker.from_env()

    def is_docker_installed(self):
        """Check if Docker is installed."""
        try:
            self.client_docker.version()
            return True
        except docker.errors.APIError:
            return False

    def is_docker_running(self):
        """Check if Docker daemon is running."""
        try:
            self.client_docker.ping()
            return True
        except docker.errors.APIError:
            return False

    def does_image_exist(self):
        """Check if a Docker image exists."""
        try:
            self.client_docker.images.get(self.image_name)
            return True
        except docker.errors.ImageNotFound:
            return False

    def does_container_exist(self):
        """Check if a Docker container exists."""
        try:
            self.client_docker.containers.get(self.container_name)
            return True
        except docker.errors.NotFound:
            return False

    def is_container_running(self):
        """Check if a Docker container is running."""
        try:
            container = self.client_docker.containers.get(self.container_name)
            return container.status == "running"
        except docker.errors.NotFound:
            return False

    def start_docker_qdrant(self):
        """Start the Qdrant Docker container.

        @note: Make sure the user has the appropriate permissions to run Docker commands
            without sudo. Otherwise, the client_docker.images.pull() command will fail.
            See U{https://docs.docker.com/engine/install/linux-postinstall/} for more details.
        """
        if not self.is_docker_installed():
            print("Docker is not installed. Please install Docker to proceed.")
            return

        if not self.is_docker_running():
            print(
                "Docker daemon is not running. Please start Docker manually and try again."
            )
            return

        if not self.does_image_exist():
            print("Image does not exist. Pulling image...")
            self.client_docker.images.pull(self.image_name)

        if not self.does_container_exist():
            print("Container does not exist. Creating container...")
            # Expand the user's home directory
            volume_path = os.path.expanduser(
                "~/.cache/" + self.container_name + "/data"
            )

            # Create the directory if it doesn't exist
            if not os.path.exists(volume_path):
                os.makedirs(volume_path)

            self.client_docker.containers.run(
                self.image_name,
                detach=True,
                name=self.container_name,
                ports={"6333/tcp": 6333},
                volumes={volume_path: {"bind": "/app/data", "mode": "rw"}},
            )
        elif not self.is_container_running():
            print("Container is not running. Starting container...")
            container = self.client_docker.containers.get(self.container_name)
            container.start()
        else:
            print("Container is already running.")

    def stop_docker_qdrant(self):
        """Stop the Qdrant Docker container."""
        if self.does_container_exist():
            container = self.client_docker.containers.get(self.container_name)
            container.stop()
            print("Stopped container")
        else:
            print("Container does not exist")


class QdrantAPI(VectorDBAPI):
    """Class to perform various Qdrant operations related to
    embeddings."""

    def __init__(self, host: str = "localhost", port: int = 6333) -> None:
        """Initialize the QdrantAPI without setting a specific
        collection.

        @type host: str
        @param host: The host address of the Qdrant server. Default is
            "localhost".
        @type port: int
        @param port: The port number of the Qdrant server. Default is
            6333.
        """
        self.client = QdrantClient(host=host, port=port)

    def create_collection(
        self,
        collection_name: str,
        properties: List[str],
        vector_size: int = 512,
    ) -> None:
        """Create a collection in Qdrant with specified properties.

        @type collection_name: str
        @param collection_name: The name of the collection.
        @type properties: List[str]
        @param properties: The list of properties for the collection.
        @type vector_size: int
        @param vector_size: The size of the embedding vectors. Default
            is 512.
        """
        self.collection_name = collection_name
        self.properties = properties
        try:
            self.client.get_collection(collection_name=self.collection_name)
            print("Collection already exists")
        except Exception:
            self.client.recreate_collection(
                collection_name=self.collection_name,
                vectors_config=VectorParams(
                    size=vector_size, distance=Distance.COSINE
                ),
            )
            print("Created new collection")

    def delete_collection(self) -> None:
        """Delete a collection in Qdrant."""
        try:
            self.client.delete_collection(collection_name=self.collection_name)
            print("Deleted collection")
        except Exception:
            print("Collection does not exist")

    def insert_embeddings(
        self,
        ids: List[str],
        embeddings: List[List[float]],
        payloads: List[Dict[str, Any]],
        batch_size: int = 50,
    ) -> None:
        """Batch insert embeddings with IDs and additional metadata into
        a collection.

        @type ids: List[str]
        @param ids: The list of instance_ids for the embeddings.
        @type embeddings: List[List[float]]
        @param embeddings: The list of embedding vectors.
        @type payloads: List[Dict[str, Any]]
        @param payloads: The list of additional metadata for the
            embeddings.
        @type batch_size: int
        @param batch_size: The batch size for inserting embeddings.
            Default is 50.
        """
        total_len = len(embeddings)

        # check if payloads key values subset of the self.properties
        for payload in payloads:
            if not set(payload.keys()).issubset(set(self.properties)):
                raise ValueError(
                    "Payload keys should be subset of the properties"
                )

        for i in range(0, total_len, batch_size):
            batch_ids = ids[i : i + batch_size]
            batch_vectors = embeddings[i : i + batch_size]
            batch_payloads = payloads[i : i + batch_size]

            batch = models.Batch(
                ids=batch_ids, vectors=batch_vectors, payloads=batch_payloads
            )

            # Upsert the batch of points to the Qdrant collection
            self.client.upsert(
                collection_name=self.collection_name, points=batch
            )

    def search_similar_embeddings(
        self, embedding: List[float], top_k: int = 5
    ) -> Tuple[List[str], List[float]]:
        """Search for the top similar embeddings in a Qdrant collection.

        @type embedding: List[float]
        @param embedding: The query embedding vector.
        @type top_k: int
        @param top_k: The number of similar embeddings to retrieve.
            Default is 5.
        @rtype: Tuple[List[str], List[float]]
        @return: The list of instance_ids of the similar embeddings and
            the list of similarity scores.
        """
        search_results = self.client.search(
            collection_name=self.collection_name,
            query_vector=embedding,
            limit=top_k,
        )

        ids, scores = [], []
        for hit in search_results:
            ids.append(hit.id)
            scores.append(hit.score)

        return ids, scores

    def get_similarity_scores(
        self,
        reference_id: str,
        other_ids: List[str],
        sort_distances: bool = True,
    ) -> Tuple[List[str], List[float]]:
        """Get a list of similarity scores between the reference
        embedding and other embeddings.

        @type reference_id: int
        @param reference_id: The instance_id of the reference embedding.
        @type other_ids: List[int]
        @param other_ids: The list of instance_ids of other embeddings
            to compare with the reference.
        @type sort_distances: bool
        @param sort_distances: Whether to sort the results by distance
            or keep the original order.
        @rtype: Tuple[List[int], List[float]
        @return: The list of instance_ids of the other embeddings and
            the list of similarity scores.
        """
        # Retrieve the embedding vector for the reference_id
        reference_embedding = self.get_embeddings_from_ids([reference_id])[0]

        # Search for similar embeddings using the reference embedding
        hits = self.client.search(
            collection_name=self.collection_name,
            query_vector=reference_embedding,
            query_filter=models.Filter(
                should=[
                    models.FieldCondition(
                        key="instance_id", match=models.MatchText(text=str(id))
                    )
                    for id in other_ids
                ]
            ),
        )

        if sort_distances:
            ids = []
            scores = []
            for hit in hits:
                ids.append(hit.id)
                scores.append(hit.score)
        else:
            ids = other_ids
            scores = [0 for _ in range(len(other_ids))]
            for hit in hits:
                i = other_ids.index(hit.id)
                scores[i] = hit.score

        return ids, scores

    def compute_similarity_matrix(self) -> List[List[float]]:
        """Compute a full similarity matrix for all embeddings in a
        Qdrant collection.

        @rtype: Tuple[List[str], List[List[float]]]
        @return: The list of instance_ids of the embeddings and the
            similarity matrix.
        @note: This method is not recommended for large collections. It
            is better to use the L{get_all_embeddings} method and
            compute the similarity matrix yourself.
        """
        # Get all embeddings
        ids, embeddings = self.retrieve_all_embeddings()
        print("Retrieved {} embeddings".format(len(embeddings)))

        # Create a list of search requests
        search_queries = [
            SearchRequest(
                vector=emb,
                with_payload=False,
                with_vector=False,
                limit=len(embeddings),
            )
            for emb in embeddings
        ]
        print("Created {} search queries".format(len(search_queries)))

        # Perform batch search
        batch_search_results = []
        for patch in range(0, len(search_queries), 100):
            batch_search_results_i = self.client.search_batch(
                collection_name=self.collection_name,
                requests=search_queries[patch : patch + 100],
            )
            batch_search_results.extend(batch_search_results_i)
            print(
                "Completed search for batch {}-{}".format(patch, patch + 100)
            )

        # Create a dictionary for O(1) lookup of ids
        id_to_index = {id: index for index, id in enumerate(ids)}

        # Get the similarity matrix
        sim_matrix = [
            [0 for _ in range(len(embeddings))] for _ in range(len(embeddings))
        ]
        for i, res in enumerate(batch_search_results):
            for hit in res:
                j = id_to_index[hit.id]
                sim_matrix[i][j] = hit.score

        print("Created similarity matrix")
        return sim_matrix

    def retrieve_embeddings_by_ids(self, ids: List[str]) -> List[List[float]]:
        """Retrieve embeddings associated with a list of IDs from a
        Qdrant collection. The order of the embeddings IS preserved.

        @type ids: List[str]
        @param ids: The list of instance_ids of the embeddings to
            retrieve.
        @rtype: List[List[float]]
        @return: The list of embedding vectors.
        """
        # Retrieve the embeddings for the given ids
        hits = self.client.retrieve(
            collection_name=self.collection_name,
            ids=ids,
            with_payload=False,
            with_vectors=True,
        )

        # Order the embeddings according to the given ids
        id_to_index = {id: index for index, id in enumerate(ids)}
        embeddings = [None for _ in range(len(ids))]
        for hit in hits:
            i = id_to_index[hit.id]
            embeddings[i] = hit.vector

        return embeddings

    def retrieve_all_ids(self) -> List[str]:
        """Retrieve all IDs from a Qdrant collection.

        @rtype: List[str]
        @return: The list of instance_ids of the embeddings.
        """
        # Get the number of points in the collection
        collection_info = self.client.get_collection(
            collection_name=self.collection_name
        )
        collection_size = collection_info.vectors_count

        if collection_size == 0:
            return []

        # Use qdrant scroll method
        hits, _offset = self.client.scroll(
            collection_name=self.collection_name, limit=collection_size
        )
        ids = [hit.id for hit in hits]
        return ids

    def retrieve_all_embeddings(self) -> Tuple[List[str], List[List[float]]]:
        """Retrieve all embeddings and their IDs from a Qdrant
        collection.

        @rtype: Tuple[List[str], List[List[float]]]
        @return: The list of instance_ids of the embeddings and the list
            of embedding vectors.
        """
        # Get the number of points in the collection
        collection_info = self.client.get_collection(
            collection_name=self.collection_name
        )
        collection_size = collection_info.vectors_count

        # Use qdrant scroll method
        hits, _offset = self.client.scroll(
            collection_name=self.collection_name,
            limit=collection_size,
            with_vectors=True,
        )

        ids, embeddings = [], []
        for hit in hits:
            ids.append(hit.id)
            embeddings.append(hit.vector)

        return ids, embeddings

    def retrieve_payloads_by_ids(
        self, ids: List[str], properties: Optional[List[str]] = None
    ) -> List[Dict[str, Any]]:
        """Retrieve specified payload properties for a list of IDs from
        a collection. The order of the labels IS preserved.

        @type ids: List[str]
        @param ids: The list of instance_ids of the embeddings to
            retrieve.
        @type properties: Optional[List[str]]
        @param properties: The list of payload properties to retrieve.
            Default is None.
        @rtype: List[Dict[str, Any]]
        @return: The list of payload dictionaries.
        """
        # Retrieve the payloads for the given ids
        hits = self.client.retrieve(
            collection_name=self.collection_name,
            ids=ids,
            with_payload=True,
            with_vectors=False,
        )

        # Order the payloads according to the given ids
        id_to_index = {id: index for index, id in enumerate(ids)}
        payloads = [None for _ in range(len(ids))]
        for hit in hits:
            i = id_to_index[hit.id]
            if properties is not None:
                payload = {prop: hit.payload[prop] for prop in properties}
            else:
                payload = hit.payload
            payloads[i] = payload

        return payloads



import json
import logging
from pathlib import Path
from typing import List, Optional, Tuple

import cv2
import numpy as np
import rich.box
import typer
import yaml
from rich import print
from rich.console import Console, group
from rich.panel import Panel
from rich.rule import Rule
from rich.table import Table
from typing_extensions import Annotated

from luxonis_ml.data import (
    Augmentations,
    LabelType,
    LuxonisDataset,
    LuxonisLoader,
    LuxonisParser,
)
from luxonis_ml.data.utils.visualizations import visualize
from luxonis_ml.enums import DatasetType

logger = logging.getLogger(__name__)

app = typer.Typer()


def complete_dataset_name(incomplete: str):
    datasets = LuxonisDataset.list_datasets()
    return [d for d in datasets if d.startswith(incomplete)]


DatasetNameArgument = Annotated[
    str,
    typer.Argument(
        ..., help="Name of the dataset.", autocompletion=complete_dataset_name
    ),
]


def check_exists(name: str):
    if not LuxonisDataset.exists(name):
        print(f"[red]Dataset [magenta]'{name}'[red] does not exist.")
        raise typer.Exit()


def get_dataset_info(name: str) -> Tuple[int, List[str], List[str]]:
    dataset = LuxonisDataset(name)
    size = len(dataset)
    classes, _ = dataset.get_classes()
    return size, classes, dataset.get_tasks()


def print_info(name: str) -> None:
    dataset = LuxonisDataset(name)
    _, classes = dataset.get_classes()
    table = Table(
        title="Classes", box=rich.box.ROUNDED, row_styles=["yellow", "cyan"]
    )
    table.add_column("Task", header_style="magenta i", max_width=30)
    table.add_column("Class Names", header_style="magenta i", max_width=50)
    for task, c in classes.items():
        table.add_row(task, ", ".join(c))

    splits = dataset.get_splits()

    @group()
    def get_sizes_panel():
        if splits is not None:
            for split, files in splits.items():
                yield f"[magenta b]{split}: [not b cyan]{len(files)}"
        else:
            yield "[red]No splits found"
        yield Rule()
        yield f"[magenta b]Total: [not b cyan]{len(dataset)}"

    @group()
    def get_panels():
        yield f"[magenta b]Name: [not b cyan]{name}"
        yield ""
        yield Panel.fit(get_sizes_panel(), title="Split Sizes")
        yield table

    print(
        Panel.fit(
            get_panels(),
            title="Dataset Info",
        )
    )


@app.command()
def info(name: DatasetNameArgument):
    """Prints information about a dataset."""
    check_exists(name)
    print_info(name)


@app.command()
def delete(name: DatasetNameArgument):
    """Deletes a dataset."""
    check_exists(name)

    dataset = LuxonisDataset(name)
    dataset.delete_dataset()
    print(f"Dataset '{name}' deleted.")


@app.command()
def ls(
    full: Annotated[
        bool, typer.Option("--full", "-f", help="Show full information.")
    ] = False,
):
    """Lists all datasets."""
    datasets = LuxonisDataset.list_datasets()
    table = Table(
        title="Datasets" + (" - Full Table" if full else ""),
        box=rich.box.ROUNDED,
        row_styles=["yellow", "cyan"],
    )
    table.add_column("Name", header_style="magenta i")
    table.add_column("Size", header_style="magenta i")
    if full:
        table.add_column("Classes", header_style="magenta i")
        table.add_column("Tasks", header_style="magenta i")
    for name in datasets:
        dataset = LuxonisDataset(name)
        rows = [name]
        try:
            size = len(dataset)
        except KeyError:
            size = -1
        rows.append(str(size))
        if full:
            _, classes, tasks = get_dataset_info(name)
            rows.extend(
                [
                    ", ".join(classes) if classes else "[red]<empty>[no red]",
                    ", ".join(tasks) if tasks else "[red]<empty>[no red]",
                ]
            )
        table.add_row(*rows)
    console = Console()
    console.print(table)


@app.command()
def inspect(
    name: DatasetNameArgument,
    view: Annotated[
        Optional[List[str]],
        typer.Option(
            ...,
            "--view",
            "-v",
            help="Which splits of the dataset to inspect.",
            case_sensitive=False,
        ),
    ] = None,
    aug_config: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--aug-config",
            "-a",
            help="Path to a config defining augmentations. "
            "This can be either a json or a yaml file.",
            metavar="PATH",
        ),
    ] = None,
    size_multiplier: Annotated[
        float,
        typer.Option(
            ...,
            "--size-multiplier",
            "-s",
            help=(
                "Multiplier for the image size. "
                "By default the images are shown in their original size."
            ),
            show_default=False,
        ),
    ] = 1.0,
):
    """Inspects images and annotations in a dataset."""

    view = view or ["train"]
    dataset = LuxonisDataset(name)
    h, w, _ = LuxonisLoader(dataset, view=view)[0][0].shape
    augmentations = None

    if aug_config is not None:
        with open(aug_config) as file:
            config = (
                yaml.safe_load(file)
                if Path(aug_config).suffix == ".yaml"
                else json.load(file)
            )
        augmentations = Augmentations([h, w], config)

    if len(dataset) == 0:
        raise ValueError(f"Dataset '{name}' is empty.")

    loader = LuxonisLoader(dataset, view=view, augmentations=augmentations)
    class_names = dataset.get_classes()[1]
    for image, labels in loader:
        image = image.astype(np.uint8)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        h, w, _ = image.shape
        new_h, new_w = int(h * size_multiplier), int(w * size_multiplier)
        image = cv2.resize(image, (new_w, new_h))
        image = visualize(image, labels, class_names)
        cv2.imshow("image", image)
        if cv2.waitKey() == ord("q"):
            break


def _parse_tasks(values: Optional[List[str]]) -> List[Tuple[LabelType, str]]:
    if not values:
        return []
    result = {}
    for value in values:
        if "=" not in value:
            raise ValueError(
                f"Invalid task format: {value}. Expected 'label_type=task_name'."
            )
        k, v = value.split("=")
        if k not in LabelType.__members__.values():
            raise ValueError(f"Invalid task type: {k}")
        result[LabelType(k.strip())] = v.strip()
    return list(result.items())


@app.command()
def parse(
    dataset_dir: Annotated[
        str, typer.Argument(..., help="Path or URL to the dataset.")
    ],
    name: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--name",
            "-n",
            autocompletion=complete_dataset_name,
            help="Name of the dataset.",
            show_default=False,
        ),
    ] = None,
    dataset_type: Annotated[
        Optional[DatasetType],
        typer.Option(
            ...,
            "--type",
            "-t",
            help="Type of the dataset. If not provided, the parser will try to recognize it automatically.",
            show_default=False,
        ),
    ] = None,
    delete_existing: Annotated[
        bool,
        typer.Option(
            ...,
            "--delete",
            "-d",
            help="If an existing dataset with the same name should "
            "be deleted before parsing.",
        ),
    ] = False,
    save_dir: Annotated[
        Optional[Path],
        typer.Option(
            ...,
            "--save-dir",
            "-s",
            help="If a remote URL is provided in 'dataset_dir', the dataset will "
            "be downloaded to this directory. Otherwise, the dataset will be "
            "downloaded to the current working directory.",
            show_default=False,
        ),
    ] = None,
    task_name: Annotated[
        Optional[List[str]],
        typer.Option(
            ...,
            "--task-name",
            "-tn",
            show_default=False,
            callback=_parse_tasks,
            help="Custom task names to override the default ones. "
            "Format: 'label_type=task_name'. E.g. 'boundingbox=detection-task'.",
        ),
    ] = None,
):
    """Parses a directory with data and creates Luxonis dataset."""
    task_name = task_name or []
    parser = LuxonisParser(
        dataset_dir,
        dataset_name=name,
        dataset_type=dataset_type,
        delete_existing=delete_existing,
        save_dir=save_dir,
        task_mapping=dict(task_name),  # type: ignore
    )
    dataset = parser.parse()

    print()
    print(Rule())
    print()
    print_info(dataset.identifier)


if __name__ == "__main__":
    app()



from abc import ABC, abstractmethod
from typing import Dict, Iterator, Tuple, Type

import numpy as np
from typing_extensions import TypeAlias

from luxonis_ml.utils import AutoRegisterMeta, Registry

from ..utils.enums import LabelType

Labels: TypeAlias = Dict[str, Tuple[np.ndarray, LabelType]]
"""C{Labels} is a dictionary mappping task names to a tuple composed of
the annotation as C{np.ndarray} and its corresponding C{LabelType}"""


LuxonisLoaderOutput: TypeAlias = Tuple[np.ndarray, Labels]
"""C{LuxonisLoaderOutput} is a tuple of an image as a C{np.ndarray>} and
a dictionary of task group names and their annotations as
L{Annotations}."""

LOADERS_REGISTRY: Registry[Type["BaseLoader"]] = Registry(name="loaders")


class BaseLoader(
    ABC, metaclass=AutoRegisterMeta, registry=LOADERS_REGISTRY, register=False
):
    """Base abstract loader class.

    Enforces the L{LuxonisLoaderOutput} output label structure.
    """

    @abstractmethod
    def __len__(self) -> int:
        """Returns the length of the dataset.

        @rtype: int
        @return: Length of the dataset.
        """
        pass

    @abstractmethod
    def __getitem__(self, idx: int) -> LuxonisLoaderOutput:
        """Loads sample from dataset.

        @type idx: int
        @param idx: Index of the sample to load.
        @rtype: L{LuxonisLoaderOutput}
        @return: Sample's data in C{LuxonisLoaderOutput} format.
        """
        pass

    def __iter__(self) -> Iterator[LuxonisLoaderOutput]:
        """Iterates over the dataset.

        @rtype: Iterator[L{LuxonisLoaderOutput}]
        @return: Iterator over the dataset.
        """
        for i in range(len(self)):
            yield self[i]



import numpy as np

from luxonis_ml.data import Augmentations, LabelType


def get_img():
    return np.zeros((320, 320, 3), dtype=np.uint8)


def get_labels():
    return {
        LabelType.CLASSIFICATION: np.array([1.0]),
        LabelType.BOUNDINGBOX: np.array(
            [
                [0.0, 0.57, 0.30, 0.17, 0.25],
                [0.0, 0.39, 0.27, 0.20, 0.10],
            ]
        ),
        LabelType.KEYPOINTS: np.array(
            [
                [0.0, 0.69, 0.37, 2.0, 0.0, 0.0, 0.0, 0.68, 0.36, 2.0],
                [0.0, 0.51, 0.33, 2.0, 0.0, 0.0, 0.0, 0.50, 0.32, 2.0],
            ]
        ),
        LabelType.SEGMENTATION: np.zeros((1, 320, 320)),
    }


def test_mosaic4():
    config = {
        "name": "Mosaic4",
        "params": {
            "p": 1.0,
            "out_width": 640,
            "out_height": 640,
        },
    }
    augmentations = Augmentations([256, 256], [config])
    data = [(get_img(), get_labels()) for _ in range(4)]
    augmentations(data, nk=3, ns=1)


def test_mixup():
    config = {
        "name": "MixUp",
        "params": {
            "p": 1.0,
        },
    }
    augmentations = Augmentations([256, 256], [config])
    data = [(get_img(), get_labels()) for _ in range(2)]
    augmentations(data, nk=3, ns=1)



from typing import List, Optional

from pydantic import Field

from luxonis_ml.utils import BaseModelExtraForbid

from .config_building_blocks import HeadType, Input, Metadata, Output


class Model(BaseModelExtraForbid):
    """Class defining a single-stage model config scheme.

    @type metadata: Metadata
    @ivar metadata: Metadata object defining the model metadata.
    @type inputs: list
    @ivar inputs: List of Input objects defining the model inputs.
    @type outputs: list
    @ivar outputs: List of Output objects defining the model outputs.
    @type heads: list
    @ivar heads: List of Head objects defining the model heads. If not
        defined, we assume a raw output.
    """

    metadata: Metadata = Field(
        description="Metadata object defining the model metadata."
    )
    inputs: List[Input] = Field(
        description="List of Input objects defining the model inputs."
    )
    outputs: List[Output] = Field(
        description="List of Output objects defining the model outputs."
    )
    heads: Optional[List[HeadType]] = Field(
        description="List of Head objects defining the model heads. If not defined, we assume a raw output."
    )



import shutil
from pathlib import Path
from typing import (
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
)

import numpy as np
import polars as pl
from pycocotools import mask as mask_utils
from typing_extensions import overload

from luxonis_ml.utils.filesystem import LuxonisFileSystem, ModuleType, PathType

from .annotation import DatasetRecord
from .base_dataset import DatasetIterator


@overload
def get_file(
    fs: LuxonisFileSystem,
    remote_path: PathType,
    local_path: PathType,
    mlflow_instance: Optional[ModuleType] = ...,
    default: Literal[None] = ...,
) -> Optional[Path]:
    pass


@overload
def get_file(
    fs: LuxonisFileSystem,
    remote_path: PathType,
    local_path: PathType,
    mlflow_instance: Optional[ModuleType] = ...,
    default: PathType = ...,
) -> Path:
    pass


def get_file(
    fs: LuxonisFileSystem,
    remote_path: PathType,
    local_path: PathType,
    mlflow_instance: Optional[ModuleType] = None,
    default: Optional[PathType] = None,
) -> Optional[Path]:
    try:
        return fs.get_file(remote_path, local_path, mlflow_instance)
    except shutil.SameFileError:
        return Path(local_path, Path(remote_path).name)
    except Exception:
        return Path(default) if default is not None else None


@overload
def find_filepath_uuid(
    filepath: PathType,
    index: Optional[pl.DataFrame],
    *,
    raise_on_missing: Literal[False] = ...,
) -> Optional[str]:
    pass


@overload
def find_filepath_uuid(
    filepath: PathType,
    index: Optional[pl.DataFrame],
    *,
    raise_on_missing: Literal[True] = ...,
) -> str:
    pass


def find_filepath_uuid(
    filepath: PathType,
    index: Optional[pl.DataFrame],
    *,
    raise_on_missing: bool = False,
) -> Optional[str]:
    if index is None:
        return None

    abs_path = str(Path(filepath).absolute().resolve())
    matched = index.filter(pl.col("original_filepath") == abs_path)

    if len(matched):
        return list(matched.select("uuid"))[0][0]
    elif raise_on_missing:
        raise ValueError(f"File {abs_path} not found in index")
    return None


@overload
def get_dir(
    fs: LuxonisFileSystem,
    remote_path: PathType,
    local_dir: PathType,
    mlflow_instance: Optional[ModuleType] = ...,
    *,
    default: Literal[None] = None,
) -> Optional[Path]:
    pass


@overload
def get_dir(
    fs: LuxonisFileSystem,
    remote_path: PathType,
    local_dir: PathType,
    mlflow_instance: Optional[ModuleType] = ...,
    *,
    default: Path = ...,
) -> Path:
    pass


def get_dir(
    fs: LuxonisFileSystem,
    remote_path: PathType,
    local_dir: PathType,
    mlflow_instance: Optional[ModuleType] = None,
    *,
    default: Optional[PathType] = None,
) -> Optional[Path]:
    try:
        return fs.get_dir(remote_path, local_dir, mlflow_instance)
    except shutil.SameFileError:
        return Path(local_dir, Path(remote_path).name)
    except Exception:
        return Path(default) if default is not None else None


def _rescale_mask(
    mask: np.ndarray,
    mask_w: int,
    mask_h: int,
    x: float,
    y: float,
    w: float,
    h: float,
) -> np.ndarray:
    return mask[
        int(y * mask_h) : int((y + h) * mask_h),
        int(x * mask_w) : int((x + w) * mask_w),
    ].astype(np.uint8)


def _rescale_rle(rle: dict, x: float, y: float, w: float, h: float) -> dict:
    height, width = rle["size"]

    if isinstance(rle["counts"], list):
        rle["counts"] = "".join(map(str, rle["counts"]))

    decoded_mask = mask_utils.decode(rle)  # type: ignore

    cropped_mask = _rescale_mask(decoded_mask, width, height, x, y, w, h)

    bbox_height = int(h * height)
    bbox_width = int(w * width)

    norm_mask = cropped_mask.astype(np.uint8)
    encoded_norm_mask = mask_utils.encode(np.asfortranarray(norm_mask))

    return {
        "height": bbox_height,
        "width": bbox_width,
        "counts": encoded_norm_mask["counts"].decode("utf-8")
        if isinstance(encoded_norm_mask["counts"], bytes)
        else encoded_norm_mask["counts"],
    }


def rescale_values(
    bbox: Dict[str, float],
    ann: Union[List, Dict],
    sub_ann_key: Literal["keypoints", "segmentation"],
) -> Optional[
    Union[
        List[Tuple[float, float, int]],
        List[Tuple[float, float]],
        Dict[str, Union[int, List[int]]],
        np.ndarray,
    ]
]:
    """Rescale annotation values based on the bounding box
    coordinates."""
    x, y, w, h = bbox["x"], bbox["y"], bbox["w"], bbox["h"]

    if sub_ann_key == "keypoints":
        return [
            (
                float(kp[0] * w + x),
                float(kp[1] * h + y),
                int(kp[2]),
            )
            for kp in ann
        ]

    if sub_ann_key == "segmentation":
        assert isinstance(ann, dict)
        if "polylines" in ann:
            return [
                (poly[0] * w + x, poly[1] * h + y) for poly in ann["polylines"]
            ]

        if "rle" in ann:
            return _rescale_rle(ann["rle"], x, y, w, h)

        if "mask" in ann:
            mask = ann["mask"]
            width, height = mask.shape
            return _rescale_mask(ann["mask"], width, height, x, y, w, h)

        raise ValueError(
            "Invalid segmentation format. Must be either 'polylines', 'rle', or 'mask'"
        )

    return None


def add_generator_wrapper(generator: DatasetIterator) -> DatasetIterator:
    """Generator wrapper to rescale and reformat annotations for each
    record in the input generator."""

    def create_new_record(
        record: Dict[str, Union[str, Dict]],
        annotation: Dict[str, Union[str, int, float, List, Dict]],
    ) -> Dict[str, Union[str, Dict]]:
        """Create a new record with the updated annotation."""
        return {
            "file": record["file"],
            "annotation": annotation,
        }

    for record in generator:
        if isinstance(record, DatasetRecord) or "annotation" not in record:
            yield record
            continue

        ann = record["annotation"]
        if ann["type"] != "detection":
            yield record
            continue

        bbox = ann.get("boundingbox", None)
        for sub_ann_key in ["boundingbox", "segmentation", "keypoints"]:
            if sub_ann_key not in ann:
                continue

            sub_ann = ann[sub_ann_key]
            if sub_ann_key == "boundingbox":
                bbox = sub_ann

            if ann.get("scaled_to_boxes", False):
                sub_ann = rescale_values(bbox, sub_ann, sub_ann_key)  # type: ignore

            task = ann.get("task", "detection")

            new_ann = {
                "type": sub_ann_key,
                "class": ann["class"],
                "task": f"{task}-{sub_ann_key}",
            }

            if sub_ann_key == "boundingbox":
                new_ann.update({"instance_id": ann["instance_id"], **bbox})
            elif sub_ann_key == "segmentation":
                if isinstance(sub_ann, list):
                    new_ann.update({"points": sub_ann, "type": "polyline"})
                elif isinstance(sub_ann, dict) and "counts" in sub_ann:
                    new_ann.update(
                        {
                            "height": sub_ann["height"],
                            "width": sub_ann["width"],
                            "counts": sub_ann["counts"],
                            "type": "rle",
                        }
                    )
            elif sub_ann_key == "keypoints":
                new_ann.update(
                    {"instance_id": ann["instance_id"], "keypoints": sub_ann}
                )

            yield create_new_record(record, new_ann)



import shutil
import tarfile
from pathlib import Path
from typing import Any, Dict, Literal

import onnx
import pytest
from heads import (
    classification_head,
    custom_segmentation_head,
    ssd_object_detection_head,
    yolo_instance_seg_kpts_head,
    yolo_instance_segmentation_head,
    yolo_keypoint_detection_head,
    yolo_obb_detection_head,
    yolo_object_detection_head,
)
from onnx import checker, helper
from onnx.onnx_pb import TensorProto
from pydantic import ValidationError

from luxonis_ml.nn_archive import ArchiveGenerator, is_nn_archive
from luxonis_ml.nn_archive.model import HeadType, Input, Output

DATA_DIR = Path("tests/data/test_nn_archive")


def create_onnx_model():
    input0 = helper.make_tensor_value_info(
        "input0", TensorProto.FLOAT, [1, 3, 64, 64]
    )
    input1 = helper.make_tensor_value_info(
        "input1", TensorProto.FLOAT, [1, 3, 128, 128]
    )

    output0 = helper.make_tensor_value_info(
        "output0", TensorProto.FLOAT, [1, 10]
    )
    output1 = helper.make_tensor_value_info(
        "output1", TensorProto.FLOAT, [1, 5, 5, 5]
    )
    graph = helper.make_graph(
        [], "DummyModel", [input0, input1], [output0, output1]
    )

    model = helper.make_model(graph, producer_name="DummyModelProducer")
    checker.check_model(model)
    onnx.save(model, str(DATA_DIR / "test_model.onnx"))


@pytest.fixture(autouse=True, scope="session")
def setup():
    DATA_DIR.mkdir(parents=True, exist_ok=True)
    create_onnx_model()

    yield

    shutil.rmtree(DATA_DIR)


@pytest.mark.parametrize("compression", ["xz", "gz", "bz2"])
@pytest.mark.parametrize(
    "head, archive_name",
    [
        (classification_head, "classification"),
        (ssd_object_detection_head, "ssd_detection"),
        (yolo_object_detection_head, "yolo_detection"),
        (yolo_instance_segmentation_head, "yolo_instance_segmentation"),
        (yolo_keypoint_detection_head, "yolo_keypoint_detection"),
        (yolo_obb_detection_head, "yolo_obb_detection"),
        (yolo_instance_seg_kpts_head, "yolo_instance_seg_kpts"),
        (custom_segmentation_head, "custom_segmentation"),
    ],
)
@pytest.mark.dependency(name="test_archive_generator")
def test_archive_generator(
    compression: Literal["xz", "gz", "bz2"],
    head: Dict[str, Any],
    archive_name: Literal[
        "classification",
        "ssd_detection",
        "yolo_detection",
        "yolo_instance_segmentation",
        "yolo_keypoint_detection",
        "yolo_obb_detection",
        "yolo_instance_seg_kpts",
        "custom_segmentation",
    ],
):
    generator = ArchiveGenerator(
        archive_name=archive_name,
        save_path="tests/data/test_nn_archive",
        cfg_dict={
            "config_version": "1.0",
            "model": {
                "metadata": {
                    "name": "test_model",
                    "path": "test_model.onnx",
                },
                "inputs": [
                    {
                        "name": "input",
                        "shape": [1, 3, 224, 224],
                        "input_type": "image",
                        "layout": "nchw",
                        "dtype": "float32",
                        "preprocessing": {
                            "mean": [0.485, 0.456, 0.406],
                            "scale": [0.229, 0.224, 0.225],
                            "reverse_channels": False,
                            "interleaved_to_planar": False,
                        },
                    }
                ],
                "outputs": [
                    {
                        "name": "output",
                        "dtype": "float32",
                    }
                ],
                "heads": [head],
            },
        },
        executables_paths=[str(DATA_DIR / "test_model.onnx")],
        compression=compression,
    )
    generator.make_archive()
    assert (DATA_DIR / "{}.tar.xz".format(archive_name)).exists()
    assert tarfile.is_tarfile(DATA_DIR / f"{archive_name}.tar.{compression}")
    with tarfile.open(DATA_DIR / f"{archive_name}.tar.{compression}") as tar:
        assert "test_model.onnx" in tar.getnames()
        assert "config.json" in tar.getnames()


def test_config_version():
    from luxonis_ml.nn_archive import Config

    cfg_dict = {
        "config_version": "1.0",
        "model": {
            "metadata": {
                "name": "test_model",
                "path": "test_model.onnx",
            },
            "inputs": [
                {
                    "name": "input",
                    "shape": [1, 3, 224, 224],
                    "input_type": "image",
                    "layout": "nchw",
                    "dtype": "float32",
                    "preprocessing": {
                        "mean": [0.485, 0.456, 0.406],
                        "scale": [0.229, 0.224, 0.225],
                        "reverse_channels": False,
                        "interleaved_to_planar": False,
                    },
                }
            ],
            "outputs": [
                {
                    "name": "output",
                    "dtype": "float32",
                }
            ],
            "heads": [],
        },
    }
    Config(**cfg_dict)
    cfg_dict["config_version"] = "1.2"
    Config(**cfg_dict)
    cfg_dict["config_version"] = "1.2.2"
    with pytest.raises(ValidationError):
        Config(**cfg_dict)
    cfg_dict["config_version"] = "1.a"
    with pytest.raises(ValidationError):
        Config(**cfg_dict)


def test_optional_head_name():
    from luxonis_ml.nn_archive.config_building_blocks.base_models.head_metadata import (
        HeadMetadata,
    )

    # without head name
    HeadType(parser="Parser", metadata=HeadMetadata(), outputs=["output"])  # type: ignore
    # with head name
    HeadType(
        parser="Parser",
        name="HeadName",
        metadata=HeadMetadata(),  # type: ignore
        outputs=["output"],
    )


def test_layout():
    default = {
        "name": "input",
        "dtype": "float32",
        "input_type": "image",
        "preprocessing": {},
    }
    inp = Input(
        **{
            **default,
            "shape": [1, 3, 224, 224],
            "layout": "nchw",
        }
    )
    assert inp.layout == "NCHW"
    inp = Input(
        **{
            **default,
            "shape": [3, 256, 256, 16],
            "layout": "chwd",
        }
    )
    assert inp.layout == "CHWD"
    out = Output(
        **{
            "name": "output",
            "dtype": "float32",
            "shape": [1, 10],
            "layout": "nc",
        }
    )
    assert out.layout == "NC"

    with pytest.raises(ValidationError):
        Input(
            **{
                **default,
                "shape": [3, 256, 256, 16],
                "layout": "1chwc2",
            }
        )

    with pytest.raises(ValidationError):
        Input(
            **{
                **default,
                "shape": [1, 3, 256, 256],
                "layout": "nch",
            }
        )

    with pytest.raises(ValidationError):
        Input(
            **{
                **default,
                "shape": [1, 3, 256, 256],
                "layout": "nchh",
            }
        )

    with pytest.raises(ValidationError):
        Output(
            **{
                "name": "output",
                "dtype": "float32",
                "shape": [1, 10],
                "layout": "ncn",
            }
        )

    with pytest.raises(ValidationError):
        Output(
            **{
                "name": "output",
                "dtype": "float32",
                "layout": list("nc"),
            }
        )


@pytest.mark.parametrize(
    "archive_name",
    [
        "classification",
        "ssd_detection",
        "yolo_detection",
        "yolo_instance_segmentation",
        "yolo_keypoint_detection",
        "yolo_obb_detection",
        "yolo_instance_seg_kpts",
        "custom_segmentation",
    ],
)
@pytest.mark.dependency(depends=["test_archive_generator"])
def test_is_nn_archive(
    archive_name: Literal[
        "classification",
        "ssd_detection",
        "yolo_detection",
        "yolo_instance_segmentation",
        "yolo_keypoint_detection",
        "yolo_obb_detection",
        "yolo_instance_seg_kpts",
        "custom_segmentation",
    ],
):
    assert is_nn_archive(DATA_DIR / f"{archive_name}.tar.xz")
    assert not is_nn_archive(DATA_DIR)
    assert not is_nn_archive(DATA_DIR / "test_model.onnx")



"""Embeddings Extractor and Storage for ONNX Models.

This module provides utility functions specifically for extracting embeddings from ONNX models,
reading images from Luxonis Filesystem (LFS), and storing/retrieving embeddings to/from disk.

Key Functions:
   - extract_embeddings(image_paths, ort_session, lfs, ...)
     Extracts embeddings from an ONNX model, reading images from LFS and handling potential errors.

   - get_image_tensors_from_LFS(image_paths, preprocess_function, lfs)
     Reads images from LFS, applies preprocessing, and returns a list of tensors.

   - preprocess_image_cv2(img)
     Example preprocessing function for resizing, normalization, and channel arrangement.

Workflow:
    1. Load the ONNX model into an ONNX Runtime session.
    2. Provide a list of image paths (stored on LFS) to the C{extract_embeddings} function.
    3. Optionally, specify a custom preprocessing function for image preparation.
    4. The function extracts embeddings in batches, handles errors, and returns the extracted embeddings.

Additional Features:
    - Saving and loading embeddings to/from disk (functions not yet implemented in this version).

Dependencies:
   - onnxruntime
   - cv2
   - numpy

Note:
    - Ensure the output_layer_name in C{extract_embeddings} matches the appropriate output layer in the ONNX model.
    - This module specifically focuses on ONNX models and reading images from Luxonis Filesystem.
"""

from typing import Callable, List, Tuple

import cv2
import numpy as np
import onnxruntime as ort

from luxonis_ml.utils import LuxonisFileSystem


def preprocess_image_cv2(img: np.ndarray) -> np.ndarray:
    """Custom preprocessing function provided by the user."""
    # Resize and convert BGR to RGB
    img = cv2.resize(img, (224, 224))
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

    # Normalize the image
    img = img.astype(np.float32) / 255.0
    mean = np.array([0.485, 0.456, 0.406])
    std = np.array([0.229, 0.224, 0.225])
    img_normalized = (img - mean) / std

    # Change to channel-first format
    img_transposed = img_normalized.transpose(2, 0, 1)

    return img_transposed


def get_image_tensors_from_remote(
    image_paths: List[str],
    preprocess_function: Callable[[np.ndarray], np.ndarray],
    lfs: LuxonisFileSystem,
) -> Tuple[np.ndarray, List[int]]:
    tensors = []
    successful_ixs = []

    for i, path in enumerate(image_paths):
        try:
            # Read image data into a byte buffer
            buffer = lfs.read_to_byte_buffer(remote_path=path).getvalue()
            image = np.frombuffer(buffer, dtype=np.uint8)
            image = cv2.imdecode(image, cv2.IMREAD_COLOR)
            processed_image = preprocess_function(image)
            tensors.append(processed_image)
            successful_ixs.append(i)
        except Exception as e:
            print(f"Error occurred while processing image: {path}, Error: {e}")
            continue

    return np.stack(tensors), successful_ixs


def extract_embeddings(
    image_paths: List[str],
    ort_session: ort.InferenceSession,
    lfs: LuxonisFileSystem,
    preprocess_function: Callable[[np.ndarray], np.ndarray],
    output_layer_name: str = "/Flatten_output_0",
    batch_size: int = 64,
) -> Tuple[List[List[float]], List[int]]:
    embeddings = []
    successful_ixs = []

    if preprocess_function is None:
        preprocess_function = preprocess_image_cv2

    for i in range(0, len(image_paths), batch_size):
        batch_paths = image_paths[i : i + batch_size]
        batch_tensors, batch_ixs = get_image_tensors_from_remote(
            batch_paths, preprocess_function, lfs
        )

        # Prepare inputs for ONNX runtime
        ort_inputs = {
            ort_session.get_inputs()[0].name: batch_tensors.astype(np.float32)
        }
        ort_outputs = ort_session.run([output_layer_name], ort_inputs)[0]

        embeddings.extend(ort_outputs.tolist())
        successful_ixs.extend([i + ix for ix in batch_ixs])

    return embeddings, successful_ixs



from typing import List, Tuple

import cv2
import onnxruntime as ort
import torch
import torch.nn as nn
import torchvision.models as models
import torchvision.models.resnet as resnet
import torchvision.transforms as transforms


# PyTorch and ONNX model loading and exporting functions
def load_model_resnet50(discard_last_layer: bool) -> nn.Module:
    """Load a pre-trained ResNet-50 model with the last fully connected
    layer removed."""
    model = models.resnet50(weights=resnet.ResNet50_Weights.IMAGENET1K_V1)
    if discard_last_layer:
        model = nn.Sequential(
            *list(model.children())[:-1]
        )  # Remove the last fully connected layer
    model.eval()
    return model


def export_model_onnx(model: nn.Module, model_path_out: str = "resnet50.onnx"):
    """Export the provided model to the ONNX format."""
    dummy_input = torch.randn(1, 3, 224, 224)

    torch.onnx.export(
        model,
        dummy_input,
        model_path_out,
        export_params=True,
        opset_version=11,
        do_constant_folding=False,
        input_names=["input"],
        output_names=["output"],
        dynamic_axes={"input": {0: "batch_size"}, "output": {0: "batch_size"}},
    )


# Embedding extraction and saving functions
def extract_embeddings_torch(
    model: torch.nn.Module, data_loader: torch.utils.data.DataLoader
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Extract embeddings from the given PyTorch model."""
    embeddings = []
    labels = []

    with torch.no_grad():
        for images, batch_labels in data_loader:
            outputs = model(images)
            embeddings.extend(outputs.squeeze())
            labels.extend(batch_labels)

    return torch.stack(embeddings), torch.tensor(labels)


def extract_embeddings_onnx(
    ort_session: ort.InferenceSession,
    data_loader: torch.utils.data.DataLoader,
    output_layer_name: str,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Extract embeddings from the given ONNX model."""
    embeddings = []
    labels = []

    with torch.no_grad():
        for images, batch_labels in data_loader:
            ort_inputs = {ort_session.get_inputs()[0].name: images.numpy()}
            outputs = ort_session.run([output_layer_name], ort_inputs)[0]
            embeddings.extend(torch.from_numpy(outputs).squeeze())
            labels.extend(batch_labels)

    return torch.stack(embeddings), torch.tensor(labels)


def save_embeddings(
    embeddings: torch.Tensor, labels: torch.Tensor, save_path: str = "./"
):
    """Save embeddings and labels tensors to the specified path."""
    torch.save(embeddings, save_path + "embeddings.pth")
    torch.save(labels, save_path + "labels.pth")


def load_embeddings(
    save_path: str = "./",
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Load embeddings and labels tensors from the specified path."""
    embeddings = torch.load(save_path + "embeddings.pth")
    labels = torch.load(save_path + "labels.pth")

    return embeddings, labels


# Image loading and preprocessing functions
def generate_new_embeddings(
    img_paths: List[str],
    ort_session: ort.InferenceSession,
    output_layer_name: str = "/Flatten_output_0",
    emb_batch_size: int = 64,
    transform: transforms.Compose = None,
):
    """Generate embeddings for new images using a given ONNX runtime
    session.

    @type img_paths: List[str]
    @param img_paths: List of image paths for new images.
    @type ort_session: L{InferenceSession}
    @param ort_session: ONNX runtime session.
    @type output_layer_name: str
    @param output_layer_name: Name of the output layer in the ONNX
        model.
    @type emb_batch_size: int
    @param emb_batch_size: Batch size for generating embeddings.
    @type transform: torchvision.transforms
    @param transform: Optional torchvision transform for preprocessing
        images.
    @rtype: List[List[float]]
    @return: List of embeddings for the new images.
    """
    # Generate embeddings for the new images using batching
    new_embeddings = []

    if transform is None:
        # Define a transformation for resizing and normalizing the images
        transform = transforms.Compose(
            [
                transforms.ToPILImage(),
                transforms.Resize(
                    (224, 224)
                ),  # Resize images to (224, 224) or any desired size
                transforms.ToTensor(),
                transforms.Normalize(
                    mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                ),  # Normalize images
            ]
        )

    for i in range(0, len(img_paths), emb_batch_size):
        batch_img_paths = img_paths[i : i + emb_batch_size]

        # Load, preprocess, and resize a batch of images
        batch_images = [cv2.imread(img_path) for img_path in batch_img_paths]
        batch_tensors = [transform(img) for img in batch_images]
        batch_tensor = torch.stack(batch_tensors).cuda()

        # Run the ONNX model on the batch
        ort_inputs = {
            ort_session.get_inputs()[0].name: batch_tensor.cpu().numpy()
        }
        ort_outputs = ort_session.run([output_layer_name], ort_inputs)

        # Append the embeddings from the batch to the new_embeddings list
        batch_embeddings = ort_outputs[0].squeeze()
        new_embeddings.extend(batch_embeddings.tolist())

    return new_embeddings



import pytest

from luxonis_ml.utils.registry import AutoRegisterMeta, Registry


@pytest.fixture(scope="function")
def registry() -> Registry:
    return Registry("test")


def test_creation():
    registry = Registry("test")
    assert registry.name == "test"
    assert not registry.module_dict
    assert len(registry) == 0
    assert str(registry) == "Registry('test')"
    assert repr(registry) == "Registry('test')"


def test_registry(registry: Registry):
    assert not registry.module_dict
    assert len(registry) == 0

    class A:
        pass

    registry.register_module(module=A)
    assert registry.get("A") is A

    @registry.register_module()
    class B:
        pass

    assert registry.get("B") is B
    assert len(registry) == 2

    registry.register_module(name="C", module=A)
    assert registry.get("C") is A
    assert len(registry) == 3

    registry.register_module(name="C", module=B, force=True)
    assert registry.get("C") is B

    with pytest.raises(KeyError):
        registry.register_module(name="C", module=A, force=False)

    @registry.register_module(name="Foo")
    class Bar:
        pass

    assert registry.get("Foo") is Bar
    with pytest.raises(KeyError):
        registry.get("Bar")


def test_autoregistry(registry: Registry):
    assert len(registry) == 0

    class Base(metaclass=AutoRegisterMeta, registry=registry, register=False):
        pass

    assert len(registry) == 0

    class A(Base):
        pass

    class B(Base):
        pass

    assert len(registry) == 2
    assert registry.get("A") is A
    assert registry.get("B") is B

    class C(Base, register_name="AliasC"):
        pass

    assert len(registry) == 3
    assert registry.get("AliasC") is C
    with pytest.raises(KeyError):
        registry.get("C")

    class D(Base, register=False):
        pass

    assert len(registry) == 3
    with pytest.raises(KeyError):
        registry.get("D")

    assert D.REGISTRY is registry

    foo_registry = Registry("foo")

    class E(Base, registry=foo_registry):
        pass

    assert len(registry) == 3
    with pytest.raises(KeyError):
        registry.get("E")

    assert len(foo_registry) == 1
    assert foo_registry.get("E") is E

    with pytest.raises(ValueError):

        class _(metaclass=AutoRegisterMeta, register=True):
            pass



import json
import logging
import random
import warnings
from collections import defaultdict
from operator import itemgetter
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np

from ..augmentations import Augmentations
from ..datasets import LuxonisDataset, load_annotation
from ..utils.enums import LabelType
from .base_loader import BaseLoader, Labels, LuxonisLoaderOutput

logger = logging.getLogger(__name__)


class LuxonisLoader(BaseLoader):
    def __init__(
        self,
        dataset: LuxonisDataset,
        view: Union[str, List[str]] = "train",
        stream: bool = False,
        augmentations: Optional[Augmentations] = None,
        *,
        force_resync: bool = False,
    ) -> None:
        """A loader class used for loading data from L{LuxonisDataset}.

        @type dataset: LuxonisDataset
        @param dataset: Instance of C{LuxonisDataset} to use.
        @type view: Union[str, List[str]]
        @param view: What splits to use. Can be either a single split or
            a list of splits. Defaults to C{"train"}.
        @type stream: bool
        @param stream: Flag for data streaming. Defaults to C{False}.
        @type augmentations: Optional[luxonis_ml.loader.Augmentations]
        @param augmentations: Augmentation class that performs
            augmentations. Defaults to C{None}.
        @type force_resync: bool
        @param force_resync: Flag to force resync from cloud. Defaults
            to C{False}.
        """

        self.logger = logging.getLogger(__name__)

        self.dataset = dataset
        self.stream = stream
        self.sync_mode = self.dataset.is_remote and not self.stream

        if self.sync_mode:
            self.dataset.sync_from_cloud(force=force_resync)

        if isinstance(view, str):
            view = [view]
        self.view = view

        df = self.dataset._load_df_offline()
        if df is None:
            raise FileNotFoundError("Cannot find dataframe")
        self.df = df

        if not self.dataset.is_remote or not self.stream:
            file_index = self.dataset._get_file_index()
            if file_index is None:
                raise FileNotFoundError("Cannot find file index")
            self.df = self.df.join(file_index, on="uuid").drop("file_right")
        else:
            raise NotImplementedError(
                "Streaming for remote bucket storage not implemented yet"
            )

        self.classes, self.classes_by_task = self.dataset.get_classes()
        self.augmentations = augmentations
        self.instances = []
        splits_path = self.dataset.metadata_path / "splits.json"
        if not splits_path.exists():
            raise RuntimeError(
                "Cannot find splits! Ensure you call dataset.make_splits()"
            )
        with open(splits_path, "r") as file:
            splits = json.load(file)

        for view in self.view:
            self.instances.extend(splits[view])

        self.idx_to_df_row: list[list[int]] = []
        for uuid in self.instances:
            boolean_mask = df["uuid"] == uuid
            row_indexes = boolean_mask.arg_true().to_list()
            self.idx_to_df_row.append(row_indexes)

        self.class_mappings = {}
        for task in df["task"].unique():
            if not task:
                continue
            class_mapping = {
                class_: i
                for i, class_ in enumerate(
                    sorted(
                        self.classes_by_task[task],
                        key=lambda x: {"background": -1}.get(x, 0),
                    )
                )
            }
            self.class_mappings[task] = class_mapping

        self.add_background = False
        # TODO: possibly check more labels
        test_image, test_labels = self._load_image_with_annotations(0)
        if LabelType.SEGMENTATION in test_labels:
            seg_masks = test_labels[LabelType.SEGMENTATION][0]
            if seg_masks.shape[0] > 1:
                unassigned_pixels = np.sum(seg_masks, axis=0) == 0

                if np.any(unassigned_pixels):
                    logger.warning(
                        "Found unassigned pixels in segmentation masks. Assigning them to `background` class (class index 0). If this is not desired then make sure all pixels are assigned to one class or rename your background class."
                    )
                    self.add_background = True
                    if (
                        "background"
                        not in self.classes_by_task[LabelType.SEGMENTATION]
                    ):
                        self.classes_by_task[LabelType.SEGMENTATION].append(
                            "background"
                        )
                        self.class_mappings[LabelType.SEGMENTATION] = {
                            class_: idx + 1
                            for class_, idx in self.class_mappings[
                                task
                            ].items()
                        }
                        self.class_mappings[LabelType.SEGMENTATION][
                            "background"
                        ] = 0

    def __len__(self) -> int:
        """Returns length of the dataset.

        @rtype: int
        @return: Length of the loader.
        """
        return len(self.instances)

    def __getitem__(self, idx: int) -> LuxonisLoaderOutput:
        """Function to load a sample consisting of an image and its
        annotations.

        @type idx: int
        @param idx: The integer index of the sample to retrieve.
        @rtype: L{LuxonisLoaderOutput}
        @return: The loader ouput consisting of the image and a
            dictionary defining its annotations.
        """

        if self.augmentations is None:
            return self._load_image_with_annotations(idx)

        indices = [idx]
        if self.augmentations.is_batched:
            if self.augmentations.aug_batch_size > len(self):
                warnings.warn(
                    f"Augmentations batch_size ({self.augmentations.aug_batch_size}) is larger than dataset size ({len(self)}), samples will include repetitions."
                )
                other_indices = [i for i in range(len(self)) if i != idx]
                picked_indices = random.choices(
                    other_indices, k=self.augmentations.aug_batch_size - 1
                )
            else:
                picked_indices = set()
                max_val = len(self)
                while (
                    len(picked_indices) < self.augmentations.aug_batch_size - 1
                ):
                    rand_idx = random.randint(0, max_val - 1)
                    if rand_idx != idx and rand_idx not in picked_indices:
                        picked_indices.add(rand_idx)
                picked_indices = list(picked_indices)

            indices.extend(picked_indices)

        out_dict: Dict[str, Tuple[np.ndarray, LabelType]] = {}
        loaded_anns = [self._load_image_with_annotations(i) for i in indices]
        random_state = random.getstate()
        np_random_state = np.random.get_state()
        if not loaded_anns[0][1]:
            img, aug_annotations = self.augmentations(
                [(loaded_anns[i][0], {}) for i in range(len(loaded_anns))],
            )

        while loaded_anns[0][1]:
            aug_input_data = []
            label_to_task = {}
            nk = 0
            ns = 0
            for img, annotations in loaded_anns:
                label_dict: Dict[LabelType, np.ndarray] = {}
                task_dict: Dict[LabelType, str] = {}
                for task in sorted(list(annotations.keys())):
                    array, label_type = annotations[task]
                    if label_type not in label_dict:
                        # ensure that bounding box annotations are added to the
                        # `label_dict` before keypoints
                        if label_type == LabelType.KEYPOINTS:
                            if (
                                LabelType.BOUNDINGBOX
                                in map(
                                    itemgetter(1), list(annotations.values())
                                )
                                and LabelType.BOUNDINGBOX not in label_dict  # type: ignore
                            ):
                                continue

                            if (
                                LabelType.BOUNDINGBOX in label_dict  # type: ignore
                                and LabelType.BOUNDINGBOX
                                in map(
                                    itemgetter(1), list(annotations.values())
                                )
                            ):
                                bbox_task = task_dict[LabelType.BOUNDINGBOX]
                                *_, bbox_suffix = bbox_task.split("-", 1)
                                *_, kp_suffix = task.split("-", 1)
                                if bbox_suffix != kp_suffix:
                                    continue

                        label_dict[label_type] = array
                        label_to_task[label_type] = task
                        task_dict[label_type] = task
                        annotations.pop(task)
                        if label_type == LabelType.KEYPOINTS:
                            nk = (array.shape[1] - 1) // 3
                        if label_type == LabelType.SEGMENTATION:
                            ns = array.shape[0]

                aug_input_data.append((img, label_dict))

            # NOTE: To ensure the same augmentation is applied to all samples
            # in case of multiple tasks per LabelType
            random.setstate(random_state)
            np.random.set_state(np_random_state)

            img, aug_annotations = self.augmentations(
                aug_input_data, nk=nk, ns=ns
            )
            for label_type, array in aug_annotations.items():
                out_dict[label_to_task[label_type]] = (array, label_type)

        return img, out_dict

    def _load_image_with_annotations(
        self, idx: int
    ) -> Tuple[np.ndarray, Labels]:
        """Loads image and its annotations based on index.

        @type idx: int
        @param idx: Index of the image
        @rtype: Tuple[np.ndarray, Labels]
        @return: Image as C{np.ndarray} in RGB format and a dictionary
            with all the present annotations
        """

        if not self.idx_to_df_row:
            raise ValueError(
                f"No data found in dataset '{self.dataset.identifier}' for {self.view} views"
            )

        ann_indices = self.idx_to_df_row[idx]

        ann_rows = [self.df.row(row) for row in ann_indices]
        if not self.dataset.is_remote:
            img_path = ann_rows[0][8]
        elif not self.stream:
            uuid = ann_rows[0][0]
            file_extension = ann_rows[0][8].rsplit(".", 1)[-1]
            img_path = self.dataset.media_path / f"{uuid}.{file_extension}"
        else:
            # TODO: add support for streaming remote storage
            raise NotImplementedError(
                "Streaming for remote bucket storage not implemented yet"
            )

        img = cv2.cvtColor(cv2.imread(str(img_path)), cv2.COLOR_BGR2RGB)
        height, width, _ = img.shape

        labels_by_task = defaultdict(list)
        instance_counters = defaultdict(int)
        for annotation_data in ann_rows:
            _, _, type_, _, class_, instance_id, task, ann_str, _ = (
                annotation_data
            )
            if instance_id < 0:
                instance_counters[task] += 1
                instance_id = instance_counters[task]
            data = json.loads(ann_str)
            if type_ == "ArrayAnnotation" and self.dataset.is_remote:
                data["path"] = self.dataset.arrays_path / data["path"]
            data.update(
                {
                    "class": class_,
                    "task": task,
                    "instance_id": instance_id,
                }
            )
            if type_ != "NoneType":
                annotation = load_annotation(type_, data)
                labels_by_task[task].append(annotation)

        labels: Labels = {}
        for task, anns in labels_by_task.items():
            assert anns, f"No annotations found for task {task}"
            anns.sort(key=lambda x: x.instance_id)
            array = anns[0].combine_to_numpy(
                anns, self.class_mappings[task], width=width, height=height
            )
            if self.add_background and task == LabelType.SEGMENTATION:
                unassigned_pixels = ~np.any(array, axis=0)
                background_idx = self.class_mappings[task]["background"]
                array[background_idx, unassigned_pixels] = 1

            labels[task] = (array, anns[0]._label_type)

        return img, labels



from typing import Final

import numpy as np
import pytest

from luxonis_ml.data.augmentations.custom.mosaic import (
    Mosaic4,
    bbox_mosaic4,
    keypoint_mosaic4,
    mosaic4,
)

WIDTH: Final[int] = 640
HEIGHT: Final[int] = 480


def test_mosaic4():
    img = (np.random.rand(HEIGHT, WIDTH, 3) * 255).astype(np.uint8)
    mosaic = mosaic4([img, img, img, img], HEIGHT, WIDTH, 0, 0, 0)
    assert mosaic.shape == (HEIGHT, WIDTH, 3)


def test_bbox_mosaic4():
    bbox = (0, 0, WIDTH, HEIGHT)
    for i in range(4):
        mosaic_bbox = bbox_mosaic4(
            bbox, HEIGHT // 2, WIDTH // 2, i, HEIGHT, WIDTH, 0, 0
        )
        assert pytest.approx(mosaic_bbox, abs=1) == (
            0,
            0,
            WIDTH // 2,
            HEIGHT // 2,
        )


def test_keypoint_mosaic4():
    for i, (w, h) in enumerate(
        [
            (WIDTH // 2, HEIGHT // 2),
            (WIDTH, HEIGHT // 2),
            (WIDTH // 2, HEIGHT),
            (WIDTH, HEIGHT),
        ]
    ):
        mosaic_keypoint = keypoint_mosaic4(
            (w, h, 0, 0), HEIGHT // 2, WIDTH // 2, i, HEIGHT, WIDTH, 0, 0
        )
        assert pytest.approx(mosaic_keypoint, abs=0.25) == (w * 2, h * 2, 0, 0)


def test_Mosaic4():
    img = (np.random.rand(HEIGHT, WIDTH, 3) * 255).astype(np.uint8)
    mosaic4 = Mosaic4(
        out_height=HEIGHT, out_width=WIDTH, always_apply=True, p=1.0
    )
    m = mosaic4(image_batch=[img, img, img, img], labels={})
    assert m["image_batch"][0].shape == (HEIGHT, WIDTH, 3)



import random
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from albumentations import BoxType, KeypointType
from albumentations.core.bbox_utils import (
    denormalize_bbox,
    normalize_bbox,
)
from albumentations.core.transforms_interface import (
    BoxInternalType,
    KeypointInternalType,
)

from ..batch_transform import BatchBasedTransform
from ..utils import AUGMENTATIONS


@AUGMENTATIONS.register_module()
class Mosaic4(BatchBasedTransform):
    def __init__(
        self,
        out_height: int,
        out_width: int,
        value: Optional[Union[int, float, List[int], List[float]]] = None,
        out_batch_size: int = 1,
        mask_value: Optional[Union[int, float, List[int], List[float]]] = None,
        always_apply: bool = False,
        p: float = 0.5,
    ):
        """Mosaic augmentation arranges selected four images into single
        image in a 2x2 grid layout. This is done in deterministic way
        meaning first image in the batch will always be in top left. The
        input images should have the same number of channels but can
        have different widths and heights. The output is cropped around
        the intersection point of the four images with the size
        (out_with x out_height). If the mosaic image is smaller than
        width x height, the gap is filled by the C{fill_value}.

        @type out_height: int
        @param out_height: Output image height. The mosaic image is
            cropped by this height around the mosaic center. If the size
            of the mosaic image is smaller than this value the gap is
            filled by the C{value}.
        @type out_width: int
        @param out_width: Output image width. The mosaic image is
            cropped by this height around the mosaic center. If the size
            of the mosaic image is smaller than this value the gap is
            filled by the C{value}.
        @type value: Optional[Union[int, float, List[int], List[float]]]
        @param value: Padding value. Defaults to C{None}.
        @type out_batch_size: int
        @param out_batch_size: Number of output images in the batch.
            Defaults to C{1}.
        @type mask_value: Optional[Union[int, float, List[int],
            List[float]]]
        @param mask_value: Padding value for masks. Defaults to C{None}.
        @type always_apply: bool
        @param always_apply: Whether to always apply the transform.
            Defaults to C{False}.
        @type p: float
        @param p: Probability of applying the transform. Defaults to
            C{0.5}.
        """

        super().__init__(batch_size=4, always_apply=always_apply, p=p)

        if out_height <= 0:
            raise ValueError(
                f"out_height should be larger than 0, got {out_height}"
            )
        if out_width <= 0:
            raise ValueError(
                f"out_width should be larger than 0, got {out_width}"
            )
        if out_batch_size <= 0:
            raise ValueError(
                f"out_batch_size should be larger than 0, got {out_batch_size}"
            )

        self.n_tiles = self.batch_size  # 4: 2x2
        self.out_height = out_height
        self.out_width = out_width
        self.value = value
        self.mask_value = mask_value
        self.out_batch_size = out_batch_size

    def get_transform_init_args_names(self) -> Tuple[str, ...]:
        """Gets the default arguments for the mixup augmentation.

        @rtype: Tuple[str, ...]
        @return: The string keywords of the arguments.
        """
        return (
            "out_height",
            "out_width",
            "replace",
            "value",
            "out_batch_size",
            "mask_value",
        )

    def _generate_random_crop_center(self) -> Tuple[int, int]:
        """Generate a random crop center within the bounds of the mosaic
        image size."""
        crop_x = random.randint(0, max(0, self.out_width))
        crop_y = random.randint(0, max(0, self.out_height))
        return crop_x, crop_y

    @property
    def targets_as_params(self):
        """List of augmentation targets.

        @rtype: List[str]
        @return: Output list of augmentation targets.
        """
        return ["image_batch"]

    def apply_to_image_batch(
        self,
        image_batch: List[np.ndarray],
        indices: List[int],
        x_crop: int,
        y_crop: int,
        **params,
    ) -> List[np.ndarray]:
        """Applies the transformation to a batch of images.

        @type image_batch: List[np.ndarray]
        @param image_batch: Batch of input images to which the
            transformation is applied.
        @type indices: List[Tuple[int, int]]
        @param indices: Indices of images in the batch.
        @type params: Any
        @param params: Additional parameters for the transformation.
        @type x_crop: int
        @param x_crop: x-coordinate of the croping start point
        @type y_crop: int
        @param y_crop: y-coordinate of the croping start point
        @rtype: List[np.ndarray]
        @return: List of transformed images.
        """
        output_batch = []
        for i_batch in range(self.out_batch_size):
            idx_chunk = indices[
                self.n_tiles * i_batch : self.n_tiles * (i_batch + 1)
            ]
            image_chunk = [image_batch[i] for i in idx_chunk]
            mosaiced = mosaic4(
                image_chunk,
                self.out_height,
                self.out_width,
                x_crop,
                y_crop,
                self.value,
            )
            output_batch.append(mosaiced)
        return output_batch

    def apply_to_mask_batch(
        self,
        mask_batch: List[np.ndarray],
        indices: List[int],
        x_crop: int,
        y_crop: int,
        **params,
    ) -> List[np.ndarray]:
        """Applies the transformation to a batch of masks.

        @type mask_batch: List[np.ndarray]
        @param mask_batch: Batch of input masks to which the
            transformation is applied.
        @type indices: List[Tuple[int, int]]
        @param indices: Indices of images in the batch.
        @type params: Any
        @param params: Additional parameters for the transformation.
        @type x_crop: int
        @param x_crop: x-coordinate of the croping start point
        @type y_crop: int
        @param y_crop: y-coordinate of the croping start point
        @rtype: List[np.ndarray]
        @return: List of transformed masks.
        """
        output_batch = []
        for i_batch in range(self.out_batch_size):
            idx_chunk = indices[
                self.n_tiles * i_batch : self.n_tiles * (i_batch + 1)
            ]
            mask_chunk = [mask_batch[i] for i in idx_chunk]
            mosaiced = mosaic4(
                mask_chunk,
                self.out_height,
                self.out_width,
                x_crop,
                y_crop,
                self.mask_value,
            )
            output_batch.append(mosaiced)
        return output_batch

    def apply_to_bboxes_batch(
        self,
        bboxes_batch: List[BoxType],
        indices: List[int],
        image_shapes: List[Tuple[int, int]],
        x_crop: int,
        y_crop: int,
        **params,
    ) -> List[BoxType]:
        """Applies the transformation to a batch of bboxes.

        @type bboxes_batch: List[BboxType]
        @param bboxes_batch: Batch of input bboxes to which the
            transformation is applied.
        @type indices: List[Tuple[int, int]]
        @param indices: Indices of images in the batch.
        @type image_shapes: List[Tuple[int, int]]
        @param image_shapes: Shapes of the input images in the batch.
        @type params: Any
        @param params: Additional parameters for the transformation.
        @type x_crop: int
        @param x_crop: x-coordinate of the croping start point
        @type y_crop: int
        @param y_crop: y-coordinate of the croping start point
        @rtype: List[BoxType]
        @return: List of transformed bboxes.
        """
        output_batch = []
        for i_batch in range(self.out_batch_size):
            idx_chunk = indices[
                self.n_tiles * i_batch : self.n_tiles * (i_batch + 1)
            ]
            bboxes_chunk = [bboxes_batch[i] for i in idx_chunk]
            shape_chunk = [image_shapes[i] for i in idx_chunk]
            new_bboxes = []
            for i in range(self.n_tiles):
                bboxes = bboxes_chunk[i]
                rows, cols = shape_chunk[i]
                for bbox in bboxes:
                    new_bbox = bbox_mosaic4(
                        bbox[:4],
                        rows,
                        cols,
                        i,
                        self.out_height,
                        self.out_width,
                        x_crop,
                        y_crop,
                    )
                    new_bboxes.append(tuple(new_bbox) + tuple(bbox[4:]))
            output_batch.append(new_bboxes)
        return output_batch

    def apply_to_keypoints_batch(
        self,
        keypoints_batch: List[KeypointType],
        indices: List[int],
        image_shapes: List[Tuple[int, int]],
        x_crop: int,
        y_crop: int,
        **params,
    ) -> List[KeypointType]:
        """Applies the transformation to a batch of keypoints.

        @type keypoints_batch: List[KeypointType]
        @param keypoints_batch: Batch of input keypoints to which the
            transformation is applied.
        @type indices: List[Tuple[int, int]]
        @param indices: Indices of images in the batch.
        @type image_shapes: List[Tuple[int, int]]
        @param image_shapes: Shapes of the input images in the batch.
        @type params: Any
        @param params: Additional parameters for the transformation.
        @type x_crop: int
        @param x_crop: x-coordinate of the croping start point
        @type y_crop: int
        @param y_crop: y-coordinate of the croping start point
        @rtype: List[KeypointType]
        @return: List of transformed keypoints.
        """
        output_batch = []
        for i_batch in range(self.out_batch_size):
            idx_chunk = indices[
                self.n_tiles * i_batch : self.n_tiles * (i_batch + 1)
            ]
            keypoints_chunk = [keypoints_batch[i] for i in idx_chunk]
            shape_chunk = [image_shapes[i] for i in idx_chunk]
            new_keypoints = []
            for i in range(self.n_tiles):
                keypoints = keypoints_chunk[i]
                rows, cols = shape_chunk[i]
                for keypoint in keypoints:
                    new_keypoint = keypoint_mosaic4(
                        keypoint[:4],
                        rows,
                        cols,
                        i,
                        self.out_height,
                        self.out_width,
                        x_crop,
                        y_crop,
                    )
                    new_keypoints.append(new_keypoint + tuple(keypoint[4:]))
            output_batch.append(new_keypoints)
        return output_batch

    def get_params_dependent_on_targets(
        self, params: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Get parameters dependent on the targets.

        @type params: Dict[str, Any]
        @param params: Dictionary containing parameters.
        @rtype: Dict[str, Any]
        @return: Dictionary containing parameters dependent on the
            targets.
        """
        image_batch = params["image_batch"]
        n = len(image_batch)
        if self.n_tiles * self.out_batch_size > n:
            raise ValueError(
                f"The batch size (= {n}) should be larger than "
                + f"{self.n_tiles} x out_batch_size (= {self.n_tiles * self.out_batch_size})"
            )
        indices = [0, 1, 2, 3]
        image_shapes = [tuple(image.shape[:2]) for image in image_batch]
        x_crop, y_crop = self._generate_random_crop_center()
        return {
            "indices": indices,
            "image_shapes": image_shapes,
            "x_crop": x_crop,
            "y_crop": y_crop,
        }


def mosaic4(
    image_batch: List[np.ndarray],
    height: int,
    width: int,
    x_crop: int,
    y_crop: int,
    value: Optional[int] = None,
) -> np.ndarray:
    """Arrange the images in a 2x2 grid layout. The input images should
    have the same number of channels but can have different widths and
    heights. The gaps are filled by the value.

    @type image_batch: List[np.ndarray]
    @param image_batch: Image list. The length should be four. Each
        image can has different size.
    @type height: int
    @param height: Height of output mosaic image
    @type width: int
    @param width: Width of output mosaic image
    @type value: Optional[int]
    @param value: Padding value
    @type x_crop: int
    @param x_crop: x-coordinate of the croping start point
    @type y_crop: int
    @param y_crop: y-coordinate of the croping start point
    @rtype: np.ndarray
    @return: Final output image
    """
    N_TILES = 4
    if len(image_batch) != N_TILES:
        raise ValueError(
            f"Length of image_batch should be 4. Got {len(image_batch)}"
        )

    for i in range(N_TILES - 1):
        if image_batch[0].shape[2:] != image_batch[i + 1].shape[2:]:
            raise ValueError(
                "All images should have the same number of channels."
            )

    if len(image_batch[0].shape) == 2:
        out_shape = [height * 2, width * 2]
    else:
        out_shape = [height * 2, width * 2, image_batch[0].shape[2]]

    dtype = image_batch[0].dtype

    img4 = np.full(
        out_shape,
        value if value is not None else 0,
        dtype=dtype,
    )

    for i, img in enumerate(image_batch):
        (h, w) = img.shape[:2]

        if i == 0:  # top left
            x1a, y1a, x2a, y2a = (
                max(width - w, 0),
                max(height - h, 0),
                width,
                height,
            )
            x1b, y1b, x2b, y2b = w - (x2a - x1a), h - (y2a - y1a), w, h
        elif i == 1:  # top right
            x1a, y1a, x2a, y2a = (
                width,
                max(height - h, 0),
                min(width + w, width * 2),
                height,
            )
            x1b, y1b, x2b, y2b = 0, h - (y2a - y1a), min(w, x2a - x1a), h
        elif i == 2:  # bottom left
            x1a, y1a, x2a, y2a = (
                max(width - w, 0),
                height,
                width,
                min(height * 2, height + h),
            )
            x1b, y1b, x2b, y2b = w - (x2a - x1a), 0, w, min(y2a - y1a, h)
        elif i == 3:  # bottom right
            x1a, y1a, x2a, y2a = (
                width,
                height,
                min(width + w, width * 2),
                min(height * 2, height + h),
            )
            x1b, y1b, x2b, y2b = 0, 0, min(w, x2a - x1a), min(y2a - y1a, h)

        img4_region = img4[y1a:y2a, x1a:x2a]
        img_region = img[y1b:y2b, x1b:x2b]

        img4_h, img4_w = img4_region.shape[:2]
        img_h, img_w = img_region.shape[:2]

        min_h = min(img4_h, img_h)
        min_w = min(img4_w, img_w)

        img4[y1a : y1a + min_h, x1a : x1a + min_w] = img[
            y1b : y1b + min_h, x1b : x1b + min_w
        ]

    img4 = img4[y_crop : y_crop + height, x_crop : x_crop + width]

    return img4


def bbox_mosaic4(
    bbox: BoxInternalType,
    rows: int,
    cols: int,
    position_index: int,
    height: int,
    width: int,
    x_crop: int,
    y_crop: int,
) -> BoxInternalType:
    """Adjust bounding box coordinates to account for mosaic grid
    position.

    This function modifies bounding boxes according to their placement
    in a 2x2 grid mosaic, shifting their coordinates based on the tile's
    relative position within the mosaic.

    @type bbox: BoxInternalType
    @param bbox: Bounding box coordinates to be transformed.
    @type rows: int
    @param rows: Height of the original image.
    @type cols: int
    @param cols: Width of the original image.
    @type position_index: int
    @param position_index: Position of the image in the 2x2 grid. (0 =
        top-left, 1 = top-right, 2 = bottom-left, 3 = bottom-right).
    @type height: int
    @param height: Height of the final output mosaic image.
    @type width: int
    @param width: Width of the final output mosaic image.
    @type x_crop: int
    @param x_crop: x-coordinate of the croping start point
    @type y_crop: int
    @param y_crop: y-coordinate of the croping start point
    @rtype: BoxInternalType
    @return: Transformed bounding box coordinates.
    """

    bbox = denormalize_bbox(bbox, rows, cols)

    if position_index == 0:
        shift_x = width - cols
        shift_y = height - rows
    elif position_index == 1:
        shift_x = width
        shift_y = height - rows
    elif position_index == 2:
        shift_x = width - cols
        shift_y = height
    elif position_index == 3:
        shift_x = width
        shift_y = height

    bbox = (
        bbox[0] + shift_x - x_crop,
        bbox[1] + shift_y - y_crop,
        bbox[2] + shift_x - x_crop,
        bbox[3] + shift_y - y_crop,
    )

    bbox = normalize_bbox(bbox, height, width)

    return bbox


def keypoint_mosaic4(
    keypoint: KeypointInternalType,
    rows: int,
    cols: int,
    position_index: int,
    height: int,
    width: int,
    x_crop: int,
    y_crop: int,
) -> KeypointInternalType:
    """Adjust keypoint coordinates based on mosaic grid position.

    This function adjusts the keypoint coordinates by placing them in
    one of the 2x2 mosaic grid cells, with shifts relative to the mosaic
    center.

    @type keypoint: KeypointInternalType
    @param keypoint: Keypoint coordinates and attributes (x, y, angle,
        scale).
    @type rows: int
    @param rows: Height of the original image.
    @type cols: int
    @param cols: Width of the original image.
    @type position_index: int
    @param position_index: Position of the image in the 2x2 grid. (0 =
        top-left, 1 = top-right, 2 = bottom-left, 3 = bottom-right).
    @type height: int
    @param height: Height of the final output mosaic image.
    @type width: int
    @param width: Width of the final output mosaic image.
    @type x_crop: int
    @param x_crop: x-coordinate of the croping start point
    @type y_crop: int
    @param y_crop: y-coordinate of the croping start point
    @rtype: KeypointInternalType
    @return: Adjusted keypoint coordinates.
    """
    x, y, angle, scale = keypoint

    if position_index == 0:
        shift_x = width - cols
        shift_y = height - rows
    elif position_index == 1:
        shift_x = width
        shift_y = height - rows
    elif position_index == 2:
        shift_x = width - cols
        shift_y = height
    elif position_index == 3:
        shift_x = width
        shift_y = height

    return x + shift_x - x_crop, y + shift_y - y_crop, angle, scale



import ast
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union

import yaml

from .filesystem import LuxonisFileSystem
from .pydantic_utils import BaseModelExtraForbid

T = TypeVar("T", bound="LuxonisConfig")


class LuxonisConfig(BaseModelExtraForbid):
    """Class for storing configuration."""

    @classmethod
    def get_config(
        cls: Type[T],
        cfg: Optional[Union[str, Dict[str, Any]]] = None,
        overrides: Optional[
            Union[Dict[str, Any], List[str], Tuple[str, ...]]
        ] = None,
    ) -> T:
        """Loads config from a yaml file or a dictionary.

        @type cfg: Optional[Union[str, dict]]
        @param cfg: Path to config file or a dictionary.
        @type overrides: Optional[Union[dict, list[str], tuple[str, ...]]]
        @param overrides: List of CLI overrides in a form of a dictionary mapping
            "dotted" keys to unparsed string or python values.
        @rtype: LuxonisConfig
        @return: Instance of the config class.
        @raise ValueError: If neither C{cfg} nor C{overrides} are provided.
        """
        if cfg is None and overrides is None:
            raise ValueError(
                "At least one of `cfg` or `overrides` must be set."
            )

        if isinstance(overrides, (list, tuple)):
            if len(overrides) % 2 != 0:
                raise ValueError(
                    "Override options should be a list of key-value pairs "
                    "but it's length is not divisible by 2."
                )

            overrides = dict(zip(overrides[::2], overrides[1::2]))

        overrides = overrides or {}
        cfg = cfg or {}

        if isinstance(cfg, str):
            fs = LuxonisFileSystem(cfg)
            buffer = fs.read_to_byte_buffer()
            data = yaml.load(buffer, Loader=yaml.SafeLoader)
        else:
            data = cfg

        cls._merge_overrides(data, overrides)
        return cls(**data)

    def __str__(self) -> str:
        return self.model_dump_json(indent=4)

    def __repr__(self) -> str:
        return self.__str__()

    def get_json_schema(self) -> Dict[str, Any]:
        """Retuns dict representation of the config json schema.

        @rtype: dict
        @return: Dictionary with config json schema.
        """
        return self.model_json_schema(mode="validation")

    def save_data(self, path: str) -> None:
        """Saves config to a yaml file.

        @type path: str
        @param path: Path to output yaml file.
        """
        with open(path, "w+") as f:
            yaml.dump(self.model_dump(), f, default_flow_style=False)

    def get(self, key_merged: str, default: Any = None) -> Any:
        """Returns a value from L{Config} based on the given key.

        If the key doesn't exist, the default value is returned.

        @type key_merged: str
        @param key_merged: Key in a form of a string with levels
            separated by dots.
        @type default: Any
        @param default: Default value to return if the key doesn't
            exist.
        @rtype: Any
        @return: Value of the key or default value.
        """
        value = self
        for key in key_merged.split("."):
            if isinstance(value, list):
                if not key.isdecimal():
                    raise ValueError(
                        f"Can't access list with non-int key `{key}`."
                    )
                index = int(key)
                if index >= len(value):
                    return default
                value = value[index]
            elif isinstance(value, dict):
                if key not in value:
                    return default
                value = value[key]
            else:
                if not hasattr(value, key):
                    return default
                value = getattr(value, key)
        return value

    @staticmethod
    def _merge_overrides(
        data: Dict[str, Any], overrides: Dict[str, Any]
    ) -> None:
        """Merges the config dictionary with the CLI overrides.

        The overrides are a dictionary mapping "dotted" keys to either
        final or unparsed values.

        @type data: dict
        @param data: Dictionary with config data.
        @type overrides: dict
        @param overrides: Dictionary with CLI overrides.
        @raise ValueError: If the overrides contain an invalid option.
        """

        def _parse_value(value: Any) -> Any:
            if not isinstance(value, str):
                return value

            try:
                return ast.literal_eval(value)
            except (ValueError, SyntaxError):
                # keep as string and hope for the best
                return value

        def _merge_recursive(
            data: Union[Dict, List], dot_name: str, value: Any
        ):
            key, *tail = dot_name.split(".")
            if not tail:
                parsed_value = _parse_value(value)
                if key.isdecimal():
                    index = int(key)
                    if not isinstance(data, list):
                        raise ValueError(
                            "int keys are not allowed for non-list values"
                        )
                    if index >= len(data):
                        data.append(parsed_value)
                    else:
                        data[index] = parsed_value
                elif isinstance(data, list):
                    raise ValueError(
                        "Only int keys are allowed for list values"
                    )
                else:
                    data[key] = parsed_value

                return

            key_tail = ".".join(tail)

            if key.isdecimal():
                index = int(key)
                if not isinstance(data, list):
                    raise ValueError(
                        "int keys are not allowed for non-list values"
                    )
                if index >= len(data):
                    index = len(data)
                    if data:
                        data.append(type(data[0])())
                        _merge_recursive(data[index], key_tail, value)
                    else:
                        # Try to guess type, backtrack if fails
                        data.append([])
                        try:
                            _merge_recursive(data[index], key_tail, value)
                        except Exception:
                            data[index] = {}
                            _merge_recursive(data[index], key_tail, value)
                else:
                    _merge_recursive(data[index], key_tail, value)
            else:
                if not isinstance(data, dict):
                    raise ValueError(
                        "Only dict values can be accessed with string keys"
                    )
                if key not in data:
                    # Try to guess type, backtrack if fails
                    data[key] = []
                    try:
                        _merge_recursive(data[key], key_tail, value)
                    except Exception:
                        data[key] = {}
                        _merge_recursive(data[key], key_tail, value)
                else:
                    _merge_recursive(data[key], key_tail, value)

        for dot_name, value in overrides.items():
            try:
                _merge_recursive(data, dot_name, value)
            except Exception as e:
                raise ValueError(f"Invalid option `{dot_name}`: {e}") from e
